Package: NeuralEstimators
Title: Likelihood-Free Parameter Estimation using Neural Networks
Version: 0.2.1
Authors@R: 
    person(given = "Matthew",
           family = "Sainsbury-Dale",
           role = c("aut", "cre"),
           email = "msainsburydale@gmail.com")
Description: An 'R' interface to the 'Julia' package 'NeuralEstimators.jl'. The package facilitates the user-friendly development of neural Bayes estimators, which are neural networks that map data to a point summary of the posterior distribution (Sainsbury-Dale et al., 2024, <doi:10.1080/00031305.2023.2249522>). These estimators are likelihood-free and amortised, in the sense that, once the neural networks are trained on simulated data, inference from observed data can be made in a fraction of the time required by conventional approaches. The package also supports amortised Bayesian or frequentist inference using neural networks that approximate the posterior or likelihood-to-evidence ratio (Zammit-Mangion et al., 2025, Sec. 3.2, 5.2, <doi:10.48550/arXiv.2404.12484>). The package accommodates any model for which simulation is feasible by allowing users to define models implicitly through simulated data.
Maintainer: Matthew Sainsbury-Dale <msainsburydale@gmail.com>
License: GPL (>= 2)
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: JuliaConnectoR, magrittr
Suggests: dplyr, ggplot2, ggplotify, ggpubr, gridExtra, knitr,
        rmarkdown, markdown, R.rsp, testthat (>= 3.0.0)
Config/testthat/edition: 3
SystemRequirements: Julia (>= 1.11)
VignetteBuilder: R.rsp
URL: https://github.com/msainsburydale/NeuralEstimators,
        https://msainsburydale.github.io/NeuralEstimators.jl/dev/
NeedsCompilation: no
Packaged: 2026-04-28 08:35:20 UTC; sainsbmd
Author: Matthew Sainsbury-Dale [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-28 10:10:02 UTC
Built: R 4.5.2; ; 2026-04-28 10:52:53 UTC; unix

# Generated by roxygen2: do not edit by hand

export(assess)
export(bias)
export(bootstrap)
export(encodedata)
export(estimate)
export(loadstate)
export(logratio)
export(plotdistribution)
export(plotestimates)
export(risk)
export(rmse)
export(sampleposterior)
export(savestate)
export(spatialgraph)
export(tanhloss)
export(train)
import(JuliaConnectoR)
importFrom(magrittr,"%>%")
importFrom(stats,reshape)
importFrom(utils,combn)

# NeuralEstimators <img align="right" width="200" src="https://github.com/msainsburydale/NeuralEstimators.jl/blob/main/docs/src/assets/logo.png?raw=true">

[![][CRAN-img]][CRAN-url]
[![][vignette-img]][vignette-url]
[![R-CMD-check](https://github.com/msainsburydale/NeuralEstimators/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/msainsburydale/NeuralEstimators/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/msainsburydale/NeuralEstimators/branch/main/graph/badge.svg?token=6cXItEsKs5)](https://app.codecov.io/gh/msainsburydale/NeuralEstimators)

[CRAN-img]: https://img.shields.io/badge/CRAN-blue.svg
[CRAN-url]: https://CRAN.R-project.org/package=NeuralEstimators

[vignette-img]: https://img.shields.io/badge/vignette-blue.svg
[vignette-url]: https://cran.r-project.org/package=NeuralEstimators/vignettes/NeuralEstimators.html

[julia-repo-img]: https://img.shields.io/badge/Julia_repo-purple.svg
[julia-repo-url]: https://github.com/msainsburydale/NeuralEstimators.jl

[julia-docs-img]: https://img.shields.io/badge/Julia_docs-purple.svg
[julia-docs-url]: https://msainsburydale.github.io/NeuralEstimators.jl/dev/

This repository contains the `R` interface to the `Julia` package `NeuralEstimators`. The package facilitates a suite of neural methods for parameter inference in scenarios where simulation from the model is feasible. These methods are **likelihood-free** and **amortised**, in the sense that, once the neural networks are trained on simulated data, they enable rapid inference across arbitrarily many observed data sets in a fraction of the time required by conventional approaches. The package caters for any model for which simulation is feasible by allowing the user to implicitly define their model via simulated data. 

See the [Julia documentation](https://msainsburydale.github.io/NeuralEstimators.jl/dev/) or the [vignette](https://cran.r-project.org/package=NeuralEstimators/vignettes/NeuralEstimators.html) to get started!

### Installation

To install the package, please:

1. **Install required software**  
   Ensure you have both [Julia](https://julialang.org/downloads/) and [R](https://www.r-project.org/) installed on your system.

2. **Install the Julia version of `NeuralEstimators`**  
   - To install the current stable version, run the following command in your terminal:  
     ```bash
     julia -e 'using Pkg; Pkg.add("NeuralEstimators")'
     ```  
   - To install the development version, run:  
     ```bash
     julia -e 'using Pkg; Pkg.add(url="https://github.com/msainsburydale/NeuralEstimators.jl")'
     ```

3. **Install the R interface to `NeuralEstimators`**  
   - To install from [CRAN](https://CRAN.R-project.org/package=NeuralEstimators), run the following command in R:  
     ```R
     install.packages("NeuralEstimators")
     ```  
   - To install the development version, first ensure you have `devtools` installed, then run:  
     ```R
     devtools::install_github("msainsburydale/NeuralEstimators")
     ```

### Supporting and citing

This software was developed as part of academic research. If you would like to support it, please star the repository. If you use the software in your research or other activities, please use the citation information accessible with the command:

```R
citation("NeuralEstimators")
```

### Contributing

If you encounter a bug or have a suggestion, please consider [opening an issue](https://github.com/msainsburydale/NeuralEstimators/issues) or submitting a pull request. Instructions for developing vignettes can be found in [vignettes/README.md](https://github.com/msainsburydale/NeuralEstimators/blob/main/vignettes/README.md). 

### Papers using NeuralEstimators

- **Likelihood-free parameter estimation with neural Bayes estimators** [[paper]](https://doi.org/10.1080/00031305.2023.2249522) [[code]](https://github.com/msainsburydale/NeuralBayesEstimators)

- **Neural methods for amortized inference** [[paper]](https://doi.org/10.1146/annurev-statistics-112723-034123)[[code]](https://github.com/andrewzm/Amortised_Neural_Inference_Review)

- **Neural Bayes estimators for irregular spatial data using graph neural networks** [[paper]](https://doi.org/10.1080/10618600.2024.2433671)[[code]](https://github.com/msainsburydale/NeuralEstimatorsGNN)

- **Neural Bayes estimators for censored inference with peaks-over-threshold models** [[paper]](https://jmlr.org/papers/v25/23-1134.html) [[code]](https://github.com/Jbrich95/CensoredNeuralEstimators)

- **Neural parameter estimation with incomplete data** [[paper]](https://arxiv.org/abs/2501.04330)[[code]](https://github.com/msainsburydale/NeuralIncompleteData)

# NeuralEstimators <img align="right" width="200" src="https://github.com/msainsburydale/NeuralEstimators.jl/blob/main/docs/src/assets/logo.png?raw=true">

[![][CRAN-img]][CRAN-url]
[![][vignette-img]][vignette-url]
[![R-CMD-check](https://github.com/msainsburydale/NeuralEstimators/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/msainsburydale/NeuralEstimators/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/msainsburydale/NeuralEstimators/branch/main/graph/badge.svg?token=6cXItEsKs5)](https://app.codecov.io/gh/msainsburydale/NeuralEstimators)

[CRAN-img]: https://img.shields.io/badge/CRAN-blue.svg
[CRAN-url]: https://CRAN.R-project.org/package=NeuralEstimators

[vignette-img]: https://img.shields.io/badge/vignette-blue.svg
[vignette-url]: https://cran.r-project.org/package=NeuralEstimators/vignettes/NeuralEstimators.html

[julia-repo-img]: https://img.shields.io/badge/Julia_repo-purple.svg
[julia-repo-url]: https://github.com/msainsburydale/NeuralEstimators.jl

[julia-docs-img]: https://img.shields.io/badge/Julia_docs-purple.svg
[julia-docs-url]: https://msainsburydale.github.io/NeuralEstimators.jl/dev/

This repository contains the `R` interface to the `Julia` package `NeuralEstimators`. The package facilitates a suite of neural methods for parameter inference in scenarios where simulation from the model is feasible. These methods are **likelihood-free** and **amortised**, in the sense that, once the neural networks are trained on simulated data, they enable rapid inference across arbitrarily many observed data sets in a fraction of the time required by conventional approaches. The package caters for any model for which simulation is feasible by allowing the user to implicitly define their model via simulated data. 

See the [Julia documentation](https://msainsburydale.github.io/NeuralEstimators.jl/dev/) or the [vignette](https://cran.r-project.org/package=NeuralEstimators/vignettes/NeuralEstimators.html) to get started!

### Installation

To install the package, please:

1. **Install required software**  
   Ensure you have both [Julia](https://julialang.org/downloads/) and [R](https://www.r-project.org/) installed on your system.

2. **Install the Julia version of `NeuralEstimators`**  
   - To install the current stable version, run the following command in your terminal:  
     ```bash
     julia -e 'using Pkg; Pkg.add("NeuralEstimators")'
     ```  
   - To install the development version, run:  
     ```bash
     julia -e 'using Pkg; Pkg.add(url="https://github.com/msainsburydale/NeuralEstimators.jl")'
     ```

3. **Install the R interface to `NeuralEstimators`**  
   - To install from [CRAN](https://CRAN.R-project.org/package=NeuralEstimators), run the following command in R:  
     ```R
     install.packages("NeuralEstimators")
     ```  
   - To install the development version, first ensure you have `devtools` installed, then run:  
     ```R
     devtools::install_github("msainsburydale/NeuralEstimators")
     ```

### Supporting and citing

This software was developed as part of academic research. If you would like to support it, please star the repository. If you use the software in your research or other activities, please use the citation information accessible with the command:

```R
citation("NeuralEstimators")
```

### Contributing

If you encounter a bug or have a suggestion, please consider [opening an issue](https://github.com/msainsburydale/NeuralEstimators/issues) or submitting a pull request. Instructions for developing vignettes can be found in [vignettes/README.md](https://github.com/msainsburydale/NeuralEstimators/blob/main/vignettes/README.md). 

### Papers using NeuralEstimators

- **Likelihood-free parameter estimation with neural Bayes estimators** [[paper]](https://doi.org/10.1080/00031305.2023.2249522) [[code]](https://github.com/msainsburydale/NeuralBayesEstimators)

- **Neural methods for amortized inference** [[paper]](https://doi.org/10.1146/annurev-statistics-112723-034123)[[code]](https://github.com/andrewzm/Amortised_Neural_Inference_Review)

- **Neural Bayes estimators for irregular spatial data using graph neural networks** [[paper]](https://doi.org/10.1080/10618600.2024.2433671)[[code]](https://github.com/msainsburydale/NeuralEstimatorsGNN)

- **Neural Bayes estimators for censored inference with peaks-over-threshold models** [[paper]](https://jmlr.org/papers/v25/23-1134.html) [[code]](https://github.com/Jbrich95/CensoredNeuralEstimators)

- **Neural parameter estimation with incomplete data** [[paper]](https://arxiv.org/abs/2501.04330)[[code]](https://github.com/msainsburydale/NeuralIncompleteData)

- **Neural Bayes inference for complex bivariate extremal dependence models** [[paper]](https://arxiv.org/abs/2503.23156)[[code]]( https://github.com/lidiamandre/NBE_classifier_depmodels)

- **Fast likelihood-free parameter estimation for Lévy processes** [[paper]](https://arxiv.org/abs/2505.01639)

#' Plot estimates vs. true values. 
#'
#' @param df a long form data frame containing fields \code{estimator}, \code{parameter}, \code{estimate}, and \code{truth}.
#' @param parameter_labels a named vector containing parameter labels.
#' @param estimator_labels a named vector containing estimator labels.
#' @return a \code{'ggplot'} of the estimates for each parameter against the true value.
#' @export
#' @examples
#' \dontrun{
#' K <- 50
#' df <- data.frame(
#'   estimator = c("Estimator 1", "Estimator 2"), 
#'   parameter = rep(c("mu", "sigma"), each = K),
#'   truth = 1:(2*K), 
#'   estimate = 1:(2*K) + rnorm(4*K)
#' )
#' estimator_labels <- c("Estimator 1" = expression(hat(theta)[1]("·")),
#'                       "Estimator 2" = expression(hat(theta)[2]("·")))
#' parameter_labels <- c("mu" = expression(mu), "sigma" = expression(sigma))
#' 
#' plotestimates(df,  parameter_labels = parameter_labels, estimator_labels)}
plotestimates <- function(df, estimator_labels = ggplot2::waiver(), parameter_labels = NULL) {
  
  truth <- estimator <- NULL # Setting the variables to NULL first to appease CRAN checks (see https://stackoverflow.com/questions/9439256/how-can-i-handle-r-cmd-check-no-visible-binding-for-global-variable-notes-when)
  
  # Cater for the case that the user has passed in an "Assessment" object
  if (!is.data.frame(df)) {
    df <- if (!is.null(df$df)) df$df else df$estimates
  }
  
  if (is.null(parameter_labels)) {
    param_labeller <- identity
  } else {
    param_labeller <- ggplot2::label_parsed
    df <- dplyr::mutate_at(df, .vars = "parameter", .funs = factor, levels = names(parameter_labels), labels = parameter_labels)
  }
  
  ggplot2::ggplot(df) + 
    ggplot2::geom_point(ggplot2::aes(x=truth, y = estimate, colour  = estimator), alpha = 0.75) + 
    ggplot2::geom_abline(colour = "black", linetype = "dashed") +
    ggplot2::facet_wrap(~parameter, scales = "free", labeller = param_labeller, nrow = 1) +
    ggplot2::scale_colour_discrete(labels = estimator_labels) +
    ggplot2::labs(colour = "") + 
    ggplot2::theme_bw() +
    ggplot2::theme(strip.background = ggplot2::element_blank())
}

# ---- plotdistribution() ----

#' Plot the empirical sampling distribution of an estimator.
#'
#' @param df a long form data frame containing fields \code{estimator}, \code{parameter}, \code{estimate}, \code{truth}, and a column (e.g., \code{replicate}) to uniquely identify each observation.
#' @param type string indicating whether to plot kernel density estimates for each individual parameter (\code{type = "density"}) or scatter plots for all parameter pairs (\code{type = "scatter"}).
#' @param parameter_labels a named vector containing parameter labels.
#' @param estimator_labels a named vector containing estimator labels.
#' @param truth_colour the colour used to denote the true parameter value.
#' @param truth_size the size of the point used to denote the true parameter value (applicable only for \code{type = "scatter"}).
#' @param truth_line_size the size of the cross-hairs used to denote the true parameter value. If \code{NULL} (default), the cross-hairs are not plotted. (applicable only for \code{type = "scatter"}).
#' @param pairs logical; should we combine the scatter plots into a single pairs plot (applicable only for \code{type = "scatter"})?
#' @param upper_triangle_plots an optional list of plots to include in the uppertriangle of the pairs plot.
#' @param legend Flag; should we include the legend (only applies when constructing a pairs plot)
#' @param return_list Flag; should the parameters be split into a list?
#' @param flip Flag; should the boxplots be "flipped" using `coord_flip()` (default `FALSE`)?
#' @return a list of \code{'ggplot'} objects or, if \code{pairs = TRUE}, a single \code{'ggplot'}.
#' @export
#' @examples
#' \dontrun{
#' # In the following, we have two estimators and, for each parameter, 50 estimates
#' # from each estimator.
#' 
#' estimators <- c("Estimator 1", "Estimator 2")
#' estimator_labels <- c("Estimator 1" = expression(hat(theta)[1]("·")),
#'                       "Estimator 2" = expression(hat(theta)[2]("·")))
#'
#' # Single parameter:
#' df <- data.frame(
#'   estimator = estimators, truth = 0, parameter = "mu",
#'   estimate  = rnorm(2*50),
#'   replicate = rep(1:50, each = 2)
#' )
#' parameter_labels <- c("mu" = expression(mu))
#' plotdistribution(df)
#' plotdistribution(df, type = "density")
#' plotdistribution(df, parameter_labels = parameter_labels, estimator_labels = estimator_labels)
#'
#' # Two parameters:
#' df <- rbind(df, data.frame(
#'   estimator = estimators, truth = 1, parameter = "sigma",
#'   estimate  = rgamma(2*50, shape = 1, rate = 1),
#'   replicate = rep(1:50, each = 2)
#' ))
#' parameter_labels <- c(parameter_labels, "sigma" = expression(sigma))
#' plotdistribution(df, parameter_labels = parameter_labels)
#' plotdistribution(df, parameter_labels = parameter_labels, type = "density")
#' plotdistribution(df, parameter_labels = parameter_labels, type = "scatter")
#'
#' # Three parameters:
#' df <- rbind(df, data.frame(
#'   estimator = estimators, truth = 0.25, parameter = "alpha",
#'   estimate  = 0.5 * runif(2*50),
#'   replicate = rep(1:50, each = 2)
#' ))
#' parameter_labels <- c(parameter_labels, "alpha" = expression(alpha))
#' plotdistribution(df, parameter_labels = parameter_labels)
#' plotdistribution(df, parameter_labels = parameter_labels, type = "density")
#' plotdistribution(df, parameter_labels = parameter_labels, type = "scatter")
#' plotdistribution(df, parameter_labels = parameter_labels, type = "scatter", pairs = TRUE)
#'
#' # Pairs plot with user-specified plots in the upper triangle:
#' upper_triangle_plots <- lapply(1:3, function(i) {
#'   x = rnorm(10)
#'   y = rnorm(10)
#'   shape = sample(c("Class 1", "Class 2"), 10, replace = TRUE)
#'   ggplot() +
#'     geom_point(aes(x = x, y = y, shape = shape)) + 
#'     labs(shape = "") +
#'     theme_bw()
#' })
#' plotdistribution(
#'     df, 
#'     parameter_labels = parameter_labels, estimator_labels = estimator_labels,
#'     type = "scatter", pairs = TRUE, upper_triangle_plots = upper_triangle_plots
#'     )}
plotdistribution <- function(
  df,
  type = c( "box", "density", "scatter"),
  parameter_labels = NULL,
  estimator_labels = ggplot2::waiver(),
  truth_colour = "black",
  truth_size = 8,
  truth_line_size = NULL,
  pairs = FALSE,
  upper_triangle_plots = NULL,
  legend = TRUE,
  return_list = FALSE, 
  flip = FALSE
  ) {
  
  # Cater for the case that the user has passed in an "Assessment" object
  if (!is.data.frame(df)) {
    df <- if (!is.null(df$df)) df$df else df$estimates
  }
  

  type <- match.arg(type)
  if(!is.logical(pairs)) stop("pairs should be logical")
  if(!is.logical(flip)) stop("flip should be logical")
  if (!all(c("estimator", "parameter", "estimate", "truth") %in% names(df))) stop("df must contain the fields estimator, parameter, estimate, and truth.")
  if ("k" %in% names(df) && length(unique(df$k)) > 1) stop("df contains a column 'k' which has more than one unique value; this means that you are trying to visualise the distribution for more than one parameter configuration. To do this, please split the data frame by k and then use lapply() to generate a list of plots grouped by k.")
  if ("m" %in% names(df) && length(unique(df$m)) > 1) stop("df contains a column 'm' which has more than one unique value; this means that you are trying to visualise the distribution for more than one sample size. To do this, please split the data frame by m and then use lapply() to generate a list of plots grouped by m.") #TODO just average over m, and inform the user that this is what we're doing? 
  if (!is.null(upper_triangle_plots) && !pairs) warning("The argument upper_triangle_plots is ignored when pairs == FALSE")

  if(is.null(parameter_labels)) {
    parameter_labels <- unique(df$parameter)
    names(parameter_labels) <- parameter_labels
  }
  p <- length(parameter_labels)
  param_names <- unique(df$parameter)
  if (!all(param_names %in% names(parameter_labels))) stop("Some parameters have not been given parameter labels: Please ensure all(unique(df$parameter) %in% names(parameter_labels))")
  if (p != length(param_names)) {
    parameter_labels <- parameter_labels[param_names]
  }

  if (p == 1 && type == "scatter") {
    warning("Setting type = 'density' since the number of parameters is equal to 1.")
    type = "density"
  }

  if (p < 3 && pairs) {
    warning("Setting pairs = FALSE since the number of parameters is less than 3.")
    pairs = FALSE
  }

  if (type == "box" | type == "density") {
    if (return_list) {
      gg <- .marginalplotlist(df, parameter_labels = parameter_labels, estimator_labels = estimator_labels, truth_colour = truth_colour, type = type, flip = flip)
    } else {
      gg <- .marginalplot(df, parameter_labels = parameter_labels, estimator_labels = estimator_labels, truth_colour = truth_colour, type = type, flip = flip)
    }
  } else if (type == "scatter") {
    gg <- .scatterplot(df, parameter_labels = parameter_labels, estimator_labels = estimator_labels, truth_colour = truth_colour, truth_size = truth_size, truth_line_size = truth_line_size)
    if (pairs) {
      gg <- .pairsplot(gg, parameter_labels = parameter_labels, upper_triangle_plots = upper_triangle_plots, legend = legend)
    }
  }

  return(gg)
}

aes_string_quiet <- function(...) suppressWarnings(ggplot2::aes_string(...))

.scatterplot <- function(df, parameter_labels, truth_colour, estimator_labels, truth_size, truth_line_size) {
  
  parameter <- NULL # Setting the variables to NULL first to appease CRAN checks (see https://stackoverflow.com/questions/9439256/how-can-i-handle-r-cmd-check-no-visible-binding-for-global-variable-notes-when)

  # all parameter pairs
  combinations <- parameter_labels %>% names %>% utils::combn(2) %>% as.matrix

  # convert to wide form (using base R to avoid dependency on tidyr)
  df <- reshape(
    df,
    timevar = "parameter",
    idvar = setdiff(names(df), c("parameter", "estimate", "truth")),
    direction = "wide"
  )
  names(df) <- gsub("(estimate|truth)\\.", "\\1_", names(df)) # rename columns to match the output of pivot_wider
  # NB the above reshaping is equivalent to:
  # df <- df %>%
  #   tidyr::pivot_wider(names_from = parameter, values_from = c("estimate", "truth")) %>%
  #   as.data.frame

  # Generate the scatterplot estimation panels
  scatterplots <- apply(combinations, 2, function(p) {

    gg <- ggplot2::ggplot(data = df[sample(nrow(df)), ]) +
      ggplot2::geom_point(
        aes_string_quiet(
          x = paste("estimate", p[1], sep = "_"),
          y = paste("estimate", p[2], sep = "_"),
          colour = "estimator"
          ),
        alpha = 0.75) +
      ggplot2::geom_point(
        aes_string_quiet(
          x = paste("truth", p[1], sep = "_"),
          y = paste("truth", p[2], sep = "_")
          ),
        colour = truth_colour, shape = "+", size = truth_size
        ) +
      ggplot2::labs(colour = "", x = parameter_labels[[p[1]]], y = parameter_labels[[p[2]]]) +
      ggplot2::scale_colour_discrete(labels = estimator_labels) +
      ggplot2::theme_bw()

    if (!is.null(truth_line_size)) {
      gg <- gg +
        ggplot2::geom_vline(aes_string_quiet(xintercept = paste("truth", p[1], sep = "_")), colour = truth_colour, size = truth_line_size) +
        ggplot2::geom_hline(aes_string_quiet(yintercept = paste("truth", p[2], sep = "_")), colour = truth_colour, size = truth_line_size)
    }

    return(gg)
  })

  return(scatterplots)
}

.marginalplot <- function(df, parameter_labels, truth_colour, type, estimator_labels, flip) {
  
  estimator <- truth <- NULL # Setting the variables to NULL first to appease CRAN checks

  if (is.null(parameter_labels)) {
    param_labeller <- identity
  } else {
    param_labeller <- ggplot2::label_parsed
    df <- dplyr::mutate_at(df, .vars = "parameter", .funs = factor, levels = names(parameter_labels), labels = parameter_labels)
  }

  gg <- ggplot2::ggplot(df)

  if (type == "box") {
    gg <- gg +
      ggplot2::geom_boxplot(ggplot2::aes(y = estimate, x = estimator, colour = estimator)) +
      ggplot2::geom_hline(ggplot2::aes(yintercept = truth), colour = truth_colour, linetype = "dashed")
  } else if (type == "density"){
    gg <- gg +
      ggplot2::geom_line(ggplot2::aes(x = estimate, group = estimator, colour = estimator), stat = "density") +
      ggplot2::geom_vline(ggplot2::aes(xintercept = truth), colour = truth_colour, linetype = "dashed")
  }

  gg <- gg +
    ggplot2::facet_wrap(parameter ~ ., scales = "free", labeller = param_labeller) +
    ggplot2::labs(colour = "") +
    ggplot2::scale_colour_discrete(labels = estimator_labels) +
    ggplot2::theme_bw() +
    ggplot2::theme(
      legend.text = ggplot2::element_text(hjust = 0),
      panel.grid = ggplot2::element_blank(),
      strip.background = ggplot2::element_blank()
      )

  if (type == "box") {
    if (flip) {
      gg <- gg +
        ggplot2::coord_flip() + 
        ggplot2::theme(axis.text.y = ggplot2::element_blank(), axis.ticks.y = ggplot2::element_blank()) + 
        ggplot2::scale_x_discrete(limits=rev) + 
        ggplot2::theme(axis.title.y = ggplot2::element_blank())
    } else {
      gg <- gg + ggplot2::theme(
        axis.text.x = ggplot2::element_blank(),
        axis.ticks.x = ggplot2::element_blank(),
        axis.title.x = ggplot2::element_blank()
      ) 
    }
  }

  return(gg)
}

#TODO can I write the following in terms of .marginalplot() to reduce code repetition?
.marginalplotlist <- function(df, parameter_labels, truth_colour, type, estimator_labels, flip) {
  
  parameter <- estimator <- truth <- NULL # Setting the variables to NULL first to appease CRAN checks

  parameters <- names(parameter_labels)

  lapply(parameters, function(param) {

    df <- dplyr::filter(df, parameter == param)
    gg <- ggplot2::ggplot(df)

    if (type == "box") {
      gg <- gg +
        ggplot2::geom_boxplot(ggplot2::aes(y = estimate, x = estimator, colour = estimator)) +
        ggplot2::geom_hline(ggplot2::aes(yintercept = truth), colour = truth_colour, linetype = "dashed")
    } else if (type == "density"){
      gg <- gg +
        ggplot2::geom_line(ggplot2::aes(x = estimate, group = estimator, colour = estimator), stat = "density") +
        ggplot2::geom_vline(ggplot2::aes(xintercept = truth), colour = truth_colour, linetype = "dashed")
    }

    gg <- gg +
      ggplot2::labs(colour = "") +
      ggplot2::scale_colour_discrete(labels = estimator_labels) +
      ggplot2::theme_bw() +
      ggplot2::theme(
        legend.text = ggplot2::element_text(hjust = 0),
        panel.grid = ggplot2::element_blank(),
        strip.background = ggplot2::element_blank()
      )

    if (type == "box") {
      gg <- gg + ggplot2::labs(x = parameter_labels[param])
      if (flip) {
        gg <- gg +
          ggplot2::coord_flip() + 
          ggplot2::theme(axis.text.y = ggplot2::element_blank(), axis.ticks.y = ggplot2::element_blank()) + 
          ggplot2::scale_x_discrete(limits=rev)
      } else {
        gg <- gg +
          ggplot2::theme(
            axis.text.x = ggplot2::element_blank(), 
            axis.ticks.x = ggplot2::element_blank()
            ) 
      }

    } else if (type == "density") {
      gg <- gg +
        ggplot2::labs(title = parameter_labels[param]) +
        ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0.5))
    }

    gg
  })
}


.pairsplot <- function(scatterplots, parameter_labels, upper_triangle_plots, legend = legend) {

  p <- length(parameter_labels)

  scatterplots <- lapply(scatterplots, function(gg) gg + ggplot2::theme(axis.title = ggplot2::element_blank()))

  # Extract legend so that it can be placed in the final plot
  scatterplot_legend.grob <- ggpubr::get_legend(scatterplots[[1]])
  scatterplots <- lapply(scatterplots, function(gg) gg + ggplot2::theme(legend.position = "none"))

  param_names <- names(parameter_labels)
  layout      <- matrix(rep(NA, p^2), nrow = p)

  # Lower-diagonal part of the plot: Parameter estimates
  layout[lower.tri(layout)] <- 1:choose(p, 2)
  plotlist <- scatterplots

  # Diagonal part of the plot (parameter names)
  diag(layout) <- (choose(p, 2) + 1):(choose(p, 2) + p)
  diag_plots <- lapply(seq_along(parameter_labels), function(i) {
    ggplot2::ggplot() +
      ggplot2::annotate("text", x = 0, y = 0, label = parameter_labels[i], size = 8) +
      ggplot2::theme_void()
  })
  plotlist <- c(plotlist, diag_plots)

  # Upper-diagonal part of the plot
  upper_idx <- (choose(p + 1, 2) + 1):p^2
  layout[upper.tri(layout)] <- upper_idx
  if (is.null(upper_triangle_plots)) {
    upper_triangle_plots <- lapply(upper_idx, function(i) ggplot2::ggplot() + ggplot2::theme_void())
    upper_triangle_plots_legend_grob <- NULL
  } else {
    upper_triangle_plots_legend_grob <<- ggpubr::get_legend(upper_triangle_plots[[1]])
    if (length(upper_triangle_plots) != choose(p, 2)) stop("The number of upper_triangle_plots should be choose(p, 2), where p is the number of parameters in the statistical model")
    upper_triangle_plots <- lapply(upper_triangle_plots, function(gg) gg + ggplot2::theme(legend.position = "none"))
  }
  plotlist <- c(plotlist, upper_triangle_plots)

  # Add the scatterplots legend
  if (legend) {
    legend_part <- rep(p^2 + 1, p)
    layout   <- cbind(legend_part, layout)
    plotlist <- c(plotlist, list(scatterplot_legend.grob))
  }

  if (!is.null(upper_triangle_plots_legend_grob)) {
    # Add upper_triangle_plots legend
    legend_part <- rep(p^2 + 2, p)
    layout <- cbind(layout, legend_part)
    plotlist <- c(plotlist, list(upper_triangle_plots_legend_grob))
  }

  # Put everything together
  legend_width <- 0.5 #NB could add an argument for the relative width of the legends
  suppressWarnings(
    gg <- gridExtra::grid.arrange(
      grobs = plotlist, layout_matrix = layout,
      widths = c(
        if(legend) legend_width,
        rep(1, p),
        if(!is.null(upper_triangle_plots_legend_grob)) legend_width
        )
      )
  )

  # Convert to object of class ggplot for consistency with other return values
  gg <- ggplotify::as.ggplot(gg)
  
  return(gg)
}

testthat::skip_on_cran()

set.seed(1)

test_that("packages can be loaded properly", {
  library("NeuralEstimators")
  expect_equal(1, 1)
})

test_that("plotestimates() is working", {
  K <- 50
  df <- data.frame(
    estimator = c("Estimator 1", "Estimator 2"), 
    parameter = rep(c("mu", "sigma"), each = K),
    truth = 1:(2*K), 
    estimate = 1:(2*K) + rnorm(4*K)
  )
  estimator_labels <- c("Estimator 1" = expression(hat(theta)[1]("·")),
                        "Estimator 2" = expression(hat(theta)[2]("·")))
  parameter_labels <- c("mu" = expression(mu), "sigma" = expression(sigma))
  plotestimates(df,  parameter_labels = parameter_labels, estimator_labels)
  
  expect_equal(1, 1)
})

test_that("plotdistribution() is working", {
  
  # Single parameter:
  estimators <- c("Estimator 1", "Estimator 2")
  df <- data.frame(
    estimator = estimators, truth = 0, parameter = "mu",
    estimate  = rnorm(2*50),
    replicate = rep(1:50, each = 2)
  )
  
  parameter_labels <- c("mu" = expression(mu))
  estimator_labels <- c("Estimator 1" = expression(hat(theta)[1]("·")),
                        "Estimator 2" = expression(hat(theta)[2]("·")))
  
  plotdistribution(df, parameter_labels = parameter_labels, estimator_labels = estimator_labels)
  plotdistribution(df, parameter_labels = parameter_labels, type = "density")
  
  # Two parameters:
  df <- rbind(df, data.frame(
    estimator = estimators, truth = 1, parameter = "sigma",
    estimate  = rgamma(2*50, shape = 1, rate = 1),
    replicate = rep(1:50, each = 2)
  ))
  parameter_labels <- c(parameter_labels, "sigma" = expression(sigma))
  plotdistribution(df, return_list = TRUE)
  plotdistribution(df, parameter_labels = parameter_labels)
  plotdistribution(df, parameter_labels = parameter_labels, flip = TRUE)
  plotdistribution(df, parameter_labels = parameter_labels, flip = TRUE, return_list = TRUE)
  plotdistribution(df, parameter_labels = parameter_labels, type = "density")
  plotdistribution(df, parameter_labels = parameter_labels, type = "scatter")
  
  # Three parameters:
  df <- rbind(df, data.frame(
    estimator = estimators, truth = 0.25, parameter = "alpha",
    estimate  = 0.5 * runif(2*50),
    replicate = rep(1:50, each = 2)
  ))
  parameter_labels <- c(parameter_labels, "alpha" = expression(alpha))
  plotdistribution(df, parameter_labels = parameter_labels)
  plotdistribution(df, parameter_labels = parameter_labels, type = "density")
  plotdistribution(df, parameter_labels = parameter_labels, type = "scatter")
  #TODO why is .pairsplot this broken?
  # plotdistribution(df, parameter_labels = parameter_labels, type = "scatter", pairs = TRUE)
  # plotdistribution(df, parameter_labels = parameter_labels, type = "scatter", pairs = TRUE, legend = FALSE)
  
  # Pairs plot with user-specified plots in the upper triangle:
  upper_triangle_plots <- lapply(1:3, function(i) {
    x = rnorm(10)
    y = rnorm(10)
    shape = sample(c("Class 1", "Class 2"), 10, replace = TRUE)
    ggplot2::ggplot() +
      ggplot2::geom_point(ggplot2::aes(x = x, y = y, shape = shape)) + 
      ggplot2::labs(shape = "") +
      ggplot2::theme_bw()
  })
  # plotdistribution(df, parameter_labels = parameter_labels, type = "scatter", pairs = TRUE, upper_triangle_plots = upper_triangle_plots)
  
  
  expect_equal(1, 1)
})

#' @title encodedata
#'
#' @description For data `Z` with missing (`NA`) entries, computes an augmented data set (U, W) where W encodes the missingness pattern as an indicator vector and U is the original data Z with missing entries replaced by a fixed constant `c`.
#'
#' @param Z data containing `NA` entries
#' @param c fixed constant with which to replace `NA` entries
#' @return Augmented data set (U, W). If `Z` is provided as a list, the return type will be a `JuliaProxy` object; these objects can be indexed in the usual manner using `[[`, or converted to an R object using `juliaGet()` (note however that `juliaGet()` can be slow for large data sets). 
#' @seealso the Julia version of [`encodedata()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/miscellaneous#Utility-functions)
#' @export
#' @examples
#' \dontrun{
#' library("NeuralEstimators")
#' Z <- matrix(c(1, 2, NA, NA, 5, 6, 7, NA, 9), nrow = 3)
#' encodedata(Z)
#' encodedata(list(Z, Z))}
encodedata <- function(Z, c = 0) {
  juliaEval('using NeuralEstimators')
  if (is.list(Z)) {
    UW <- juliaLet('encodedata.(Z; c = c)', Z = Z, c = c)
  } else {
    UW <- juliaLet('encodedata(Z; c = c)', Z = Z, c = c)
  }
  return(UW)
}

#' @title tanhloss
#' @description For \code{k} > 0, defines Julia code that defines the loss function,
#' \deqn{L(\hat{\theta}, \theta) = \tanh\left(\frac{|\hat{\theta} - \theta|}{k}\right),}
#' which approximates the 0-1 loss as \code{k} tends to zero. 
#' 
#' The resulting string is intended to be used in the function \code{\link{train}}, but can also be converted to a callable function using \code{juliaEval}. 
#' @param k Positive numeric value that controls the smoothness of the approximation.
#' @return String defining the tanh loss function in Julia code.
#' @export 
tanhloss <- function(k) paste0("(x, y) -> tanhloss(x, y, ", k, ")")

#' @title spatialgraph
#'
#' @description Constructs a graph object for use in a graph neural network (GNN). 
#'
#' @param S Spatial locations, provided as:
#'   - An \eqn{n\times 2}{n x 2} matrix when locations are fixed across replicates, where \eqn{n} is the number of spatial locations. 
#'   - A list of \eqn{n_i \times 2}{ni x 2} matrices when locations vary across replicates.
#'   - A list of the above elements (i.e., a list of matrices or a list of lists of matrices) when constructing graphs from multiple data sets.
#' @param Z Spatial data, provided as:
#'   - An \eqn{n\times m}{n x m} matrix when locations are fixed, where \eqn{m} is the number of replicates.
#'   - A list of \eqn{n_i}{ni}-vectors when locations vary across replicates. 
#'   - A list of the above elements (i.e., a list of matrices or a list of lists of vectors) when constructing graphs from multiple data sets.
#' @param isotropic Logical. If `TRUE`, edge features store the spatial distance (magnitude) 
#'   between nodes. If `FALSE`, the spatial displacement or spatial location is stored, depending 
#'   on the value of `stationary`.
#' @param stationary Logical. If `TRUE`, edge features store the spatial displacement 
#'   (vector difference) between nodes, capturing both magnitude and direction. If `FALSE`, 
#'   edge features include the full spatial locations of both nodes.
#' @param ... Additional keyword arguments from the Julia function [`adjacencymatrix()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/miscellaneous#Utility-functions) 
#' that define the neighborhood of each node, with the default being a randomly 
#' selected set of `k=30` neighbors within a radius of `r=0.15` spatial distance units.
#' @export 
#' @return A `GNNGraph` (`JuliaProxy` object) or, if multiple data sets are provided, a vector of `GNNGraph` objects which can be indexed in the usual manner using `[[` or converted to an R list using a combination of indexing and `lapply`. 
#' @examples
#' \dontrun{
#' library("NeuralEstimators")
#' 
#' # Number of replicates
#' m <- 5
#' 
#' # Spatial locations fixed for all replicates
#' n <- 100 
#' S <- matrix(runif(n * 2), n, 2)
#' Z <- matrix(runif(n * m), n, m)
#' g <- spatialgraph(S, Z)
#' 
#' # Spatial locations varying between replicates
#' n <- sample(50:100, m, replace = TRUE)
#' S <- lapply(n, function(ni) matrix(runif(ni * 2), ni, 2))
#' Z <- lapply(n, function(ni) runif(ni))
#' g <- spatialgraph(S, Z)
#' 
#' # Multiple data sets: Spatial locations fixed for all replicates within a given data set
#' K <- 15 # number of data sets
#' n <- sample(50:100, K, replace = TRUE) # number of spatial locations can vary between data sets
#' S <- lapply(1:K, function(k) matrix(runif(n[k] * 2), n[k], 2))
#' Z <- lapply(1:K, function(k) matrix(runif(n[k] * m), n[k], m))
#' g <- spatialgraph(S, Z)
#' 
#' # Multiple data sets: Spatial locations varying between replicates within a given data set
#' S <- lapply(1:K, function(k) {
#'   lapply(1:m, function(i) {
#'   ni <- sample(50:100, 1)       # randomly generate the number of locations for each replicate
#'   matrix(runif(ni * 2), ni, 2)  # generate the spatial locations
#'   })
#' })
#' Z <- lapply(1:K, function(k) {
#'   lapply(1:m, function(i) {
#'     n <- nrow(S[[k]][[i]])
#'     runif(n)  
#'   })
#' })
#' g <- spatialgraph(S, Z)
#' }
spatialgraph <- function(S, Z, isotropic = TRUE, stationary = TRUE, ...) {
  juliaEval('using NeuralEstimators')
  
  # Capture additional arguments from ...
  extra_args <- list(...)
  
  # Type check and conversion for `r` and `k` if provided
  if ("r" %in% names(extra_args)) {
    if (!is.numeric(extra_args$r)) {
      stop("`r` must be a numeric value (float).")
    }
    extra_args$r <- as.numeric(extra_args$r)
  }
  
  if ("k" %in% names(extra_args)) {
    if (!is.numeric(extra_args$k) || extra_args$k %% 1 != 0) {
      stop("`k` must be an integer value.")
    }
    extra_args$k <- as.integer(extra_args$k)
  }
  
  # Prepare Julia call string for keyword arguments
  extra_args_string <- paste0(
    names(extra_args), " = ", sapply(extra_args, function(arg) {
      if (is.character(arg)) paste0('"', arg, '"') else {
        if (!is.integer(arg) && is.numeric(arg) && arg %% 1 == 0) { # Ensure float is passed as float 
          return(paste0(arg, ".0"))
        } else {
          return(arg)
        }
      }
    }), collapse = ", "
  )
  
  # Determine if we have a single data set or multiple data sets 
  if (
    is.matrix(S) && is.matrix(Z) || 
    is.list(S) && all(sapply(S, is.matrix)) && is.list(Z) && all(sapply(Z, is.vector))
    ) {
    spatialgraph_call <- "spatialgraph"
  } else {
    spatialgraph_call <- "spatialgraph."
  }
    
  # Build Julia call dynamically to include additional arguments
  julia_call <- paste0(
    spatialgraph_call, 
    "(S, Z; stationary = stationary, isotropic = isotropic",
    if (length(extra_args) > 0) paste0(", ", extra_args_string) else "",
    ")"
  )
  
  # Execute the Julia call
  g <- juliaLet(julia_call, S = S, Z = Z, stationary = stationary, isotropic = isotropic)
  
  return(g)
}

testthat::skip_on_cran()

set.seed(1)

test_that("packages can be loaded properly", {
  library("NeuralEstimators")
  library("JuliaConnectoR")
  expect_equal(1, 1)
})

test_that("julia can be called", {
  x <- juliaEval('
  1 + 1
')
  expect_equal(x, 2)
})

test_that("Flux.jl is available", {
  juliaEval('
  # Install the package if not already installed
  using Pkg
  installed = "Flux" in keys(Pkg.project().dependencies)
  if !installed
    Pkg.add("Flux")  
  end
  using Flux
')
  expect_equal(1, 1)
})

test_that("NeuralEstimators.jl is available", {
  juliaEval('
  # Install the package if not already installed
  using Pkg
  installed = "NeuralEstimators" in keys(Pkg.project().dependencies)
  if !installed
    Pkg.add(url = "https://github.com/msainsburydale/NeuralEstimators.jl") 
  end
  using NeuralEstimators
')
  expect_equal(1, 1)
})

test_that("encodedata() is working", {
  Z <- matrix(c(1, 2, NA, NA, 5, 6, 7, NA, 9), nrow = 3)
  UW <- encodedata(Z)
  UW <- encodedata(list(Z, Z))
  expect_equal(length(UW), 2)
})

#TODO why is this failing on CI but not on my computer?
# test_that("spatialgraph() is working", {
#   # Number of replicates and spatial dimension
#   m <- 5
# 
#   # Spatial locations fixed for all replicates
#   n <- 100
#   S <- matrix(runif(n * 2), n, 2)
#   Z <- matrix(runif(n * m), n, m)
#   g <- spatialgraph(S, Z)
# 
#   # Spatial locations varying between replicates
#   n <- sample(50:100, m, replace = TRUE)
#   S <- lapply(n, function(ni) matrix(runif(ni * 2), ni, 2))
#   Z <- lapply(n, function(ni) runif(ni))
#   g <- spatialgraph(S, Z)
# 
#   spatialgraph(S, Z)
#   spatialgraph(S, Z, k = 1)
#   spatialgraph(S, Z, k = 100)
#   spatialgraph(S, Z, k = 10L, r = 0.1)
#   spatialgraph(S, Z, k = 10.0, r = 0.1)
#   spatialgraph(S, Z, k = 10, r = 1.0)
# 
#   # Multiple data sets: Spatial locations fixed for all replicates within a given data set
#   K <- 15 # number of data sets
#   n <- sample(50:100, K, replace = TRUE) # number of spatial locations can vary between data sets
#   S <- lapply(1:K, function(k) matrix(runif(n[k] * 2), n[k], 2))
#   Z <- lapply(1:K, function(k) matrix(runif(n[k] * m), n[k], m))
#   g <- spatialgraph(S, Z)
#   
#   # Multiple data sets: Spatial locations varying between replicates within a given data set
#   S <- lapply(1:K, function(k) {
#     lapply(1:m, function(i) {
#       ni <- sample(50:100, 1)       # randomly generate the number of locations for each replicate
#       matrix(runif(ni * 2), ni, 2)  # generate the spatial locations
#     })
#   })
#   Z <- lapply(1:K, function(k) {
#     lapply(1:m, function(i) {
#       n <- nrow(S[[k]][[i]])
#       runif(n)  
#     })
#   })
#   g <- spatialgraph(S, Z)
# })

# Same as JuliaConnectoR::juliaSetupOk(), but also check minimum version number
.juliaSetupOk <- function(version = "1") {
  
  juliaCmd <- NULL
  try({
    # juliaCmd <- JuliaConnectoR:::getJuliaExecutablePath()
    juliaCmd <- asNamespace("JuliaConnectoR")$getJuliaExecutablePath()
  })
  if (is.null(juliaCmd)) {
    message("Julia not found")
    return(FALSE)
  }
  
  # juliaVersion <- JuliaConnectoR:::getJuliaVersionViaCmd(juliaCmd)
  juliaVersion <-  asNamespace("JuliaConnectoR")$getJuliaVersionViaCmd(juliaCmd)
  if (is.null(juliaVersion)) {
    message("Julia could not be started")
    return(FALSE)
  }
  
  # Convert versions to numeric vectors
  juliaVersionNum <- as.integer(unlist(strsplit(juliaVersion, ".", fixed = TRUE)))
  requiredVersionNum <- as.integer(unlist(strsplit(version, ".", fixed = TRUE)))
  
  # Pad versions with zeros for safe comparison (e.g., "1.11" -> c(1,11,0) if necessary)
  lengthDiff <- length(requiredVersionNum) - length(juliaVersionNum)
  if (lengthDiff > 0) {
    juliaVersionNum <- c(juliaVersionNum, rep(0, lengthDiff))
  } else if (lengthDiff < 0) {
    requiredVersionNum <- c(requiredVersionNum, rep(0, -lengthDiff))
  }
  
  # Compare versions element-wise
  if (isTRUE(all(juliaVersionNum >= requiredVersionNum))) {
    return(TRUE)
  } else {
    message("Julia version must be at least ", version)
    return(FALSE)
  }
}

.getNeuralEstimators <- function() {
  if (.juliaSetupOk() && juliaEval('"NeuralEstimators" in keys(Pkg.project().dependencies)')) {
    juliaImport("NeuralEstimators") 
  } else {
    stop("Julia package NeuralEstimators.jl not found. If it is installed but not detected, please report this issue to the package maintainer.")
  }
}

#' @title Train a neural estimator
#' 
#' @description The function caters for different variants of "on-the-fly" simulation. 
#' Specifically, a \code{sampler} can be provided to continuously sample new 
#' parameter vectors from the prior, and a \code{simulator} can be provided to 
#' continuously simulate new data conditional on the parameters. If provided 
#' with specific sets of parameters (\code{theta_train} and \code{theta_val}) 
#' and/or data (\code{Z_train} and \code{Z_val}), they will be held fixed during 
#' training.
#' 
#' Note that using \code{R} functions to perform "on-the-fly" simulation requires the user to have installed the Julia package \code{RCall}.
#'
#' @param estimator a neural estimator
#' @param sampler a function that takes an integer \code{K}, samples \code{K} parameter vectors from the prior, and returns them as a px\code{K} matrix
#' @param simulator a function that takes a px\code{K} matrix of parameters and an integer \code{m}, and returns \code{K} simulated data sets each containing \code{m} independent replicates
#' @param theta_train a set of parameters used for updating the estimator using stochastic gradient descent
#' @param theta_val a set of parameters used for monitoring the performance of the estimator during training
#' @param Z_train a simulated data set used for updating the estimator using stochastic gradient descent
#' @param Z_val a simulated data set used for monitoring the performance of the estimator during training
#' @param sampler_args a list of positional arguments passed to the parameter sampler; if provided, `sampler` is called as \code{sampler(K, sampler_args...)}.
#' @param sampler_kwargs a list of positional arguments passed to the parameter sampler; if provided, `sampler` is called as \code{sampler(K; sampler_kwargs...)}.
#' @param simulator_args a list of positional arguments passed to the data simulator; if provided, `simulator` is called as \code{simulator(theta, simulator_args...)}.
#' @param simulator_kwargs a list of positional arguments passed to the data simulator; if provided, `simulator` is called as \code{simulator(theta; simulator_kwargs...)}.
#' @param m deprecated; use \code{simulator_args}
#' @param K the number of parameter vectors sampled in the training set at each epoch; the size of the validation set is set to \code{K}/5.
#' @param loss the loss function: a string ('absolute-error' for mean-absolute-error loss or 'squared-error' for mean-squared-error loss), or a string of Julia code defining the loss function. For some classes of estimators (e.g., `PosteriorEstimator`, `QuantileEstimator`, `RatioEstimator`), the loss function does not need to be specified.
#' @param learning_rate the initial learning rate for the optimiser ADAM (default 5e-4) 
#' @param epochs the number of epochs to train the neural network. An epoch is one complete pass through the entire training data set when doing stochastic gradient descent.
#' @param stopping_epochs cease training if the risk doesn't improve in this number of epochs (default 5).
#' @param batchsize the batchsize to use when performing stochastic gradient descent, that is, the number of training samples processed between each update of the neural-network parameters. 
#' @param savepath path to save the trained estimator and other information; if null (default), nothing is saved. Otherwise, the neural-network parameters (i.e., the weights and biases) will be saved during training as `bson` files; the risk function evaluated over the training and validation sets will also be saved, in the first and second columns of `loss_per_epoch.csv`, respectively; the best parameters (as measured by validation risk) will be saved as `best_network.bson`. 
#' @param use_gpu a boolean indicating whether to use the GPU if one is available 
#' @param verbose a boolean indicating whether information, including empirical risk values and timings, should be printed to the console during training.
#' @param epochs_per_Z_refresh integer indicating how often to refresh the training data
#' @param epochs_per_theta_refresh integer indicating how often to refresh the training parameters; must be a multiple of \code{epochs_per_Z_refresh}
#' @param simulate_just_in_time  flag indicating whether we should simulate "just-in-time", in the sense that only a \code{batchsize} number of parameter vectors and corresponding data are in memory at a given time
#' @return a trained neural estimator or, if \code{m} is a vector, a list of trained neural estimators
#' @export
#' @seealso the Julia version of [`train()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/training), [assess()] for assessing an estimator post training, and [estimate()]/[sampleposterior()] for making inference with observed data
#' @examples
#' \dontrun{
#' # Construct a neural Bayes estimator for replicated univariate Gaussian 
#' # data with unknown mean and standard deviation. 
#' 
#' # Load R and Julia packages
#' library("NeuralEstimators")
#' library("JuliaConnectoR")
#' juliaEval("using NeuralEstimators, Flux")
#' 
#' # Define the neural-network architecture
#' estimator <- juliaEval('
#'  n = 1    # dimension of each replicate
#'  d = 2    # number of parameters in the model
#'  w = 32   # width of each hidden layer
#'  psi = Chain(Dense(n, w, relu), Dense(w, w, relu))
#'  phi = Chain(Dense(w, w, relu), Dense(w, d))
#'  deepset = DeepSet(psi, phi)
#'  estimator = PointEstimator(deepset)
#' ')
#' 
#' # Sampler from the prior
#' sampler <- function(K) {
#'   mu    <- rnorm(K)      # Gaussian prior for the mean
#'   sigma <- rgamma(K, 1)  # Gamma prior for the standard deviation
#'   theta <- matrix(c(mu, sigma), byrow = TRUE, ncol = K)
#'   return(theta)
#' }
#' 
#' # Data simulator
#' simulator <- function(theta_set, m) {
#'   apply(theta_set, 2, function(theta) {
#'     t(rnorm(m, theta[1], theta[2]))
#'   }, simplify = FALSE)
#' }
#' 
#' # Train using fixed parameter and data sets 
#' theta_train <- sampler(10000)
#' theta_val   <- sampler(2000)
#' m <- 30 # number of iid replicates
#' Z_train <- simulator(theta_train, m)
#' Z_val   <- simulator(theta_val, m)
#' estimator <- train(estimator, 
#'                    theta_train = theta_train, 
#'                    theta_val = theta_val, 
#'                    Z_train = Z_train, 
#'                    Z_val = Z_val)
#'                    
#' ##### Simulation on-the-fly using R functions ####
#' 
#' juliaEval("using RCall") # requires the Julia package RCall
#' estimator <- train(estimator, sampler = sampler, simulator = simulator, m = m)
#' 
#' ##### Simulation on-the-fly using Julia functions ####
#' 
#' # Defining the sampler and simulator in Julia can improve computational 
#' # efficiency by avoiding the overhead of communicating between R and Julia. 
#' 
#' juliaEval("using Distributions")
#' 
#' # Parameter sampler
#' sampler <- juliaEval("
#'       function sampler(K)
#'       	mu = rand(Normal(0, 1), K)
#'       	sigma = rand(Gamma(1), K)
#'       	theta = hcat(mu, sigma)'
#'       	return theta
#'       end")
#' 
#' # Data simulator
#' simulator <- juliaEval("
#'       function simulator(theta_matrix, m)
#'       	Z = [rand(Normal(theta[1], theta[2]), 1, m) for theta in eachcol(theta_matrix)]
#'       	return Z
#'       end")
#' 
#' # Train
#' estimator <- train(estimator, sampler = sampler, simulator = simulator, m = m)}
train <- function(estimator,
                  sampler = NULL,   
                  simulator = NULL, 
                  theta_train = NULL,
                  theta_val = NULL,
                  Z_train = NULL,
                  Z_val = NULL,
                  K = 10000,
                  m = NULL, # deprecated
                  sampler_args = NULL,
                  sampler_kwargs = NULL, 
                  simulator_args = NULL,
                  simulator_kwargs = NULL,
                  loss = "absolute-error",
                  learning_rate = 5e-4,
                  epochs = 100,
                  batchsize = 32,
                  savepath = NULL,
                  stopping_epochs = 5,
                  epochs_per_Z_refresh = 1,      
                  epochs_per_theta_refresh = 1,  
                  simulate_just_in_time = FALSE, 
                  use_gpu = TRUE,
                  verbose = TRUE
                  ) {

  # Deprecation coercion 
  if (!is.null(m)) {
    # warning("The argument `m` in `train()` is deprecated; please use `simulator_kwargs`") #TODO how should R users pass this though? Should be a NamedTuple
    simulator_args <- juliaLet('(m,)', m = as.integer(m))
  }
  
  # Convert numbers that should be integers (so that the user can write 32 rather than 32L)
  K <- as.integer(K)
  epochs <- as.integer(epochs)
  batchsize <- as.integer(batchsize)
  stopping_epochs <- as.integer(stopping_epochs)
  epochs_per_Z_refresh <- as.integer(epochs_per_Z_refresh)
  epochs_per_theta_refresh <- as.integer(epochs_per_theta_refresh)
  
  # Coerce theta_train and theta_val to 1xK matrices if given as K-vectors (which will often be the case in single-parameter settings)
  if (is.vector(theta_train)) theta_train <- t(theta_train)
  if (is.vector(theta_val)) theta_val <- t(theta_val)
  
  # logic check
  if (!is.null(sampler) && (!is.null(Z_train) || !is.null(Z_val))) stop("One cannot combine continuous resampling of the parameters through `sampler` with fixed simulated data sets, `Z_train` and `Z_val`")
  
  # Metaprogramming: Define the Julia code based on the given arguments
  train_code <- "train(estimator,"
  if (is.null(sampler)) {
    if (is.null(theta_train) || is.null(theta_val)) stop("A parameter `sampler` or sampled parameter sets `theta_train` and `theta_val` must be provided")
    train_code <- paste(train_code, "theta_train, theta_val,")
  } else {
    if (!is.null(theta_train) || !is.null(theta_val)) stop("Only one of `sampler` or `theta_train` and `theta_val` should be provided")
    train_code <- paste(train_code, "sampler,")
  }
  if (is.null(simulator)) {
    if (is.null(Z_train) || is.null(Z_val)) stop("A data `simulator` or simulated data sets `Z_train` and `Z_val` must be provided")
    train_code <- paste(train_code, "Z_train, Z_val,")
  } else {
    if (!is.null(Z_train) || !is.null(Z_val)) stop("Only one of `simulator` or `Z_train` and `Z_val` should be provided")
    train_code <- paste(train_code, "simulator,")
  }
  
  # If `sampler` and `simulator` are not Julia functions (i.e., they lack "JLFUN" 
  # attributes), then we need to define Julia functions that invoke them. We do 
  # this using RCall (see https://juliainterop.github.io/RCall.jl/stable/).
  # Since JuliaConnectoR creates a separate R environment when Julia is 
  # initialised, we must use the macro @rput to move the R functions to this 
  # separate R environment before the R functions can be invoked. 
  if (!is.null(sampler) && !("JLFUN" %in% names(attributes(sampler)))) {
    tryCatch( { juliaEval("using RCall") }, error = function(e) "using R functions to perform 'on-the-fly' simulation requires the user to have installed the Julia package RCall")
    juliaLet('using RCall; @rput sampler', sampler = sampler)
    sampler <- juliaEval('
        using RCall
        sampler(K) = rcopy(R"sampler($K)")
        sampler(K, sampler_args) = rcopy(R"sampler($K, $sampler_args)")
                         ')
  }
  if (!is.null(simulator) && !("JLFUN" %in% names(attributes(simulator)))) {
    tryCatch( { juliaEval("using RCall") }, error = function(e) "using R functions to perform 'on-the-fly' simulation requires the user to have installed the Julia package RCall")
    juliaLet('using RCall; @rput simulator', simulator = simulator)
    simulator <- juliaEval('using RCall; simulator(theta, m) = rcopy(R"simulator($theta, $m)")')
  }
  
  # Metaprogramming: All other keyword arguments for on-the-fly simulation 
  if (!is.null(simulator)) {
    if (!is.null(simulator_args))   train_code <- paste(train_code, "simulator_args = simulator_args,")
    if (!is.null(simulator_kwargs)) train_code <- paste(train_code, "simulator_kwargs = simulator_kwargs,")
    train_code <- paste(train_code, "epochs_per_Z_refresh = epochs_per_Z_refresh, simulate_just_in_time = simulate_just_in_time,")
  }
  if (!is.null(sampler)) {
    if (!is.null(sampler_args))   train_code <- paste(train_code, "sampler_args = sampler_args,")
    if (!is.null(sampler_kwargs)) train_code <- paste(train_code, "sampler_kwargs = sampler_kwargs,")
    train_code <- paste(train_code, "K = K, epochs_per_theta_refresh = epochs_per_theta_refresh,")
  }
  
  # Identify which loss function we are using; if it is a string that matches
  # absolute-error or squared-error, convert it to the Julia function
  # corresponding to those loss functions. Otherwise, pass it in unchanged,
  # so that the user can provide a Julia function defining a custom loss function.
  if (loss == "absolute-error") {
    loss = juliaEval('mae(yhat, y) = mean(abs.(yhat .- y))')
  } else if (loss == "squared-error") {
    loss = juliaEval('mse(yhat, y) = mean(abs2.(yhat .- y))')
  } else {
    loss = juliaEval(loss)
  }
  
  # Metaprogramming: load Julia packages and add keyword arguments that are applicable to all methods of train()
  code <- paste(
  "
  using NeuralEstimators
  using Optimisers: Adam
  using Statistics: mean
  
  estimator = ",
     train_code,
    "
    loss = loss,
    optimiser = Optimisers.Adam(learning_rate), 
    epochs = epochs,
    batchsize = batchsize,
    savepath = savepath,
    stopping_epochs = stopping_epochs,
    use_gpu = use_gpu,
    verbose = verbose
  )

  estimator")

  # Run the Julia code and pass the arguments from R to Julia
  estimator = juliaLet(
     code,
     estimator = estimator, 
     sampler = sampler, simulator = simulator,
     theta_train = theta_train, theta_val = theta_val, 
     Z_train = Z_train, Z_val = Z_val, 
     K = K, 
     sampler_args = sampler_args,
     sampler_kwargs = sampler_kwargs,
     simulator_args = simulator_args,
     simulator_kwargs = simulator_kwargs,
     loss = loss,
     learning_rate = learning_rate,
     epochs = epochs,
     batchsize = batchsize,
     savepath = savepath,
     stopping_epochs = stopping_epochs,
     use_gpu = use_gpu,
     verbose = verbose, 
     epochs_per_theta_refresh = epochs_per_theta_refresh,  
     epochs_per_Z_refresh = epochs_per_Z_refresh,      
     simulate_just_in_time = simulate_just_in_time
  )

  return(estimator)
}

#' @title Load a saved state of a neural estimator
#' 
#' @description Load a saved state of a neural estimator (e.g., optimised neural-network parameters). Useful for amortised inference, whereby a neural network is trained once and then used repeatedly to make inference with new data sets.
#'
#' @param estimator the neural estimator that we wish to load the state into
#' @param filename file name (including path) of the neural-network state stored in a \code{bson} file
#' @return `estimator` updated with the saved state 
#' @export
loadstate <- function(estimator, filename) {
  juliaEval('using NeuralEstimators, Flux')
  juliaEval('using BSON: @load')
  juliaLet(
    '
    @load filename model_state
    Flux.loadmodel!(estimator, model_state)
    estimator
    ',
    estimator = estimator, filename = filename
  )
}

#' @title save the state of a neural estimator
#' @param estimator the neural estimator that we wish to save
#' @param filename file in which to save the neural-network state as a \code{bson} file
#' @return No return value, called for side effects
#' @export
savestate <- function(estimator, filename) {
  juliaEval('using NeuralEstimators, Flux')
  juliaEval('using BSON: @save')
  juliaLet(
    '
    model_state = Flux.state(estimator)
    @save filename model_state
    ',
    estimator = estimator, filename = filename
  )
}

#' @title computes a Monte Carlo approximation of an estimator's Bayes risk
#' @param assessment an object returned by [assess()] 
#' @param loss a binary operator defining the loss function (default absolute-error loss)
#' @param average_over_parameters if \code{TRUE}, the loss is averaged over all parameters; otherwise (default), the loss is averaged over each parameter separately
#' @param average_over_sample_sizes if \code{TRUE} (default), the loss is averaged over all sample sizes (the column \code{m} in \code{df}); otherwise, the loss is averaged over each sample size separately
#' @return a dataframe giving an estimate of the Bayes risk
#' @seealso [assess()], [bias()], [rmse()]
#' @export
risk <- function(assessment, 
                 loss = function(x, y) abs(x - y), 
                 average_over_parameters = FALSE, 
                 average_over_sample_sizes = TRUE
                 ) {
  
  if (is.list(assessment)) df <- if (!is.null(assessment$df)) assessment$df else assessment$estimates
  if (is.data.frame(assessment)) df <- assessment
  
  truth <- NULL # Setting the variables to NULL first to appease CRAN checks (see https://stackoverflow.com/questions/9439256/how-can-i-handle-r-cmd-check-no-visible-binding-for-global-variable-notes-when)
  
  # Determine which variables we are grouping by
  grouping_variables <- intersect(
    c(
      "estimator",
      if (!average_over_parameters) "parameter",
      if (!average_over_sample_sizes) "m"
    ),
    names(df)
  )
  
  
  # Compute the risk 
  dplyr::mutate(df, loss = loss(estimate, truth)) %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(grouping_variables))) %>%
    dplyr::summarise(risk = mean(loss))
}

#' @title computes a Monte Carlo approximation of an estimator's bias
#' @inheritParams risk 
#' @param ... optional arguments inherited from `risk()` (excluding the argument `loss`)
#' @return a dataframe giving the estimated bias
#' @seealso [assess()], [risk()], [rmse()]
#' @export
bias <- function(assessment, ...) {
  df <- risk(assessment, loss = function(x, y) x - y, ...)
  names(df)[names(df) == "risk"] <- "bias"
  return(df)
}

#' @title computes a Monte Carlo approximation of an estimator's root-mean-square error (RMSE)
#' @inheritParams risk 
#' @param ... optional arguments inherited from `risk()` (excluding the argument `loss`)
#' @return a dataframe giving the estimated RMSE
#' @seealso [assess()], [bias()], [risk()]
#' @export
rmse <- function(assessment, ...) {
  df <- risk(assessment, loss = function(x, y) (x - y)^2, ...)
  df$risk <- sqrt(df$risk)
  names(df)[names(df) == "risk"] <- "rmse"
  return(df)
}

#TODO add output information for posterior samples
#' @title assess a neural estimator
#' @param estimator a neural estimator (or a list of neural estimators)
#' @param parameters true parameters, stored as a \eqn{d\times K}{dxK} matrix, where \eqn{d} is the dimension of the parameter vector and \eqn{K} is the number of sampled parameter vectors
#' @param Z data simulated conditionally on the \code{parameters}. If \code{length(Z)} > K, the parameter matrix will be recycled by horizontal concatenation as `parameters = parameters[, rep(1:K, J)]`, where `J = length(Z) / K`
#' @param ... additional keyword arguments passed to the Julia version of [`assess()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/assessment)
#' @return a list of two data frames: \code{runtimes} contains the
#' total time taken for each estimator, while \code{df} is a long-form
#' data frame with columns:
#' \itemize{
#' \item{"parameter"; the name of the parameter}
#' \item{"truth"; the true value of the parameter}
#' \item{"estimate"; the estimated value of the parameter}
#' \item{"k"; the index of the parameter vector in the test set}
#' \item{"j"; the index of the data set}
#' }
#' @seealso [risk()], [rmse()], [bias()], [plotestimates()], and [plotdistribution()] for computing various empirical diagnostics and visualisations from an object returned by `assess()`
#' @export
assess <- function(
    estimator, 
    parameters,
    Z,
    ...
) {
  
  if (is.vector(parameters)) parameters <- t(parameters)
  
  NE <- .getNeuralEstimators()
  assessment <- NE$assess(estimator, parameters, Z, ...)
  
  estimates <- juliaLet('assessment.estimates', assessment = assessment)
  runtimes  <- juliaLet('assessment.runtime', assessment = assessment)
  
  estimates <- as.data.frame(estimates)
  runtimes  <- as.data.frame(runtimes)
  output <- list(estimates = estimates, runtimes = runtimes)
  
  # Add assessment.samples if it is available
  if (juliaLet("hasproperty(assessment, :samples) && !isnothing(assessment.samples)", assessment = assessment)) {
    samples <- juliaLet('assessment.samples', assessment = assessment)
    if (!is.null(samples)) {
      #samples <- as.data.frame(samples) # NB this conversion takes a really long time, so just pro
      output <- c(output, list(samples = samples))
    }
  }
  
  return(output)
}

#' @title sampleposterior
#' 
#' @description Samples from the approximate posterior distribution given data `Z`. 
#'
#' @param estimator a neural posterior or likelihood-to-evidence-ratio estimator
#' @param Z data in a format amenable to the neural-network architecture of `estimator`
#' @param N number of approximate posterior samples to draw
#' @param ... additional keyword arguments passed to the Julia version of [`sampleposterior()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/inference#Inference-with-observed-data)
#' @return d × `N` matrix of posterior samples, where d is the dimension of the parameter vector. If `Z` contains multiple independent data sets, a list of matrices will be returned
#' @seealso [estimate()] for making inference with neural Bayes estimators
#' @export
sampleposterior <- function(estimator, Z, N = 1000, ...) {
  NE <- .getNeuralEstimators()
  NE$sampleposterior(estimator, Z, N = as.integer(N), ...)
}

#' @title estimate
#'
#' @description Apply a neural Bayes estimator to data
#'
#' @param estimator a neural estimator that can be applied to data in a call of the form `estimator(Z)`
#' @param Z data in a format amenable to the neural-network architecture of `estimator`
#' @param X additional inputs to the neural network; if provided, the call will be of the form `estimator((Z, X))`
#' @param batchsize the batch size for applying `estimator` to `Z`
#' @param ... additional keyword arguments passed to the Julia version of [`estimate()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/inference#Inference-with-observed-data)
#' @return a matrix of outputs resulting from applying `estimator` to `Z` (and possibly `X`)
#' @seealso [sampleposterior()] for making inference with neural posterior or likelihood-to-evidence-ratio estimators
#' @export
estimate <- function(estimator, Z, X = NULL, batchsize = 32, ...) {
  NE <- .getNeuralEstimators()
  if (!is.null(X)) {
    input <- juliaLet('(Z, X)', Z = Z, X = X)
  } else {
    input <- Z
  }
  
  thetahat <- NE$estimate(estimator, input, batchsize = as.integer(batchsize), ...)
  thetahat <- juliaLet('Float64.(thetahat)', thetahat = thetahat) # convert to regular matrix and Float64
  return(thetahat)
}

#' @title logratio
#'
#' @description Apply a neural ratio estimator to data
#'
#' @param estimator a neural ratio estimator
#' @param Z data in a format amenable to the neural-network architecture of `estimator`
#' @param grid matrix of parameter values, where each column is a parameter configuration
#' @param batchsize the batch size
#' @param ... additional keyword arguments passed to the Julia version of [`logratio()`](https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/inference#Inference-with-observed-data)
#' @return A matrix of log ratios with one row per data set and one column per grid point
#' @seealso [sampleposterior()] for making posterior inferences
#' @export
logratio <- function(estimator, Z, grid, batchsize = 32, ...) {
  NE <- .getNeuralEstimators()
  log_ratios <- NE$logratio(estimator, Z, grid = grid, batchsize = as.integer(batchsize), ...)
  log_ratios <- juliaLet('Float64.(log_ratios)', log_ratios = log_ratios) # convert to regular matrix and Float64
  return(log_ratios)
}

#' @title bootstrap
#' 
#' @description Compute bootstrap estimates from a neural point estimator
#'
#' @param estimator a neural point estimator
#' @param Z either a list of data sets simulated conditionally on the fitted parameters (parametric bootstrap); or a single observed data set containing independent replicates, which will be sampled with replacement `B` times (non-parametric bootstrap)
#' @param B number of non-parametric bootstrap samples
#' @param blocks integer vector specifying the blocks in non-parameteric bootstrap. For example, with 5 replicates, the first two corresponding to block 1 and the remaining three corresponding to block 2, `blocks` should be \code{c(1,1,2,2,2)}
#' @param use_gpu boolean indicating whether to use the GPU if it is available
#' @return d × `B` matrix, where d is the dimension of the parameter vector 
#' @export
bootstrap <- function(estimator,
                      Z,
                      B = 400,
                      blocks = NULL,
                      use_gpu = TRUE
) {
  
  B <- as.integer(B)
  if (!is.list(Z)) Z <- list(Z)
  
  if (length(Z) > 1) {
    #NB Just using estimate() since that is all that needs to be done here
    thetahat <- estimate(estimator, Z, use_gpu = use_gpu)
  } else {
    NE <- .getNeuralEstimators()
    thetahat <- NE$bootstrap(estimator, Z, use_gpu = use_gpu, B = B, blocks = blocks)
  }
  
  return(thetahat)
}

testthat::skip_on_cran()

set.seed(1)

test_that("packages can be loaded properly", {
  library("NeuralEstimators")
  library("JuliaConnectoR")
  expect_equal(1, 1)
})

test_that("julia can be called", {
  x <- juliaEval('
  1 + 1
')
  expect_equal(x, 2)
})

test_that("Flux.jl is available", {
  juliaEval('
  # Install the package if not already installed
  using Pkg
  installed = "Flux" in keys(Pkg.project().dependencies)
  if !installed
    Pkg.add("Flux")  
  end
  using Flux
')
  expect_equal(1, 1)
})

test_that("NeuralEstimators.jl is available", {
  juliaEval('
  # Install the package if not already installed
  using Pkg
  installed = "NeuralEstimators" in keys(Pkg.project().dependencies)
  if !installed
    Pkg.add(url = "https://github.com/msainsburydale/NeuralEstimators.jl") 
  end
  using NeuralEstimators
')
  expect_equal(1, 1)
})

test_that("Optim.jl is available", {
  juliaEval('
  # Install the package if not already installed
  using Pkg
  installed = "Optim" in keys(Pkg.project().dependencies)
  if !installed
    Pkg.add("Optim") 
  end
  using Optim
')
  expect_equal(1, 1)
})

test_that("a neural estimator can be initialised", {
  
  ## Using Flux code directly
  estimator <<- juliaEval('
  using NeuralEstimators, Flux

  p = 2    # number of parameters in the statistical model
  w = 32   # number of neurons in each layer

  psi = Chain(Dense(1, w, relu), Dense(w, w, relu), Dense(w, w, relu))
  phi = Chain(Dense(w, w, relu), Dense(w, p))
  estimator = PointEstimator(DeepSet(psi, phi))
  ')
  
  expect_equal(1, 1)
})

# Sampler from the prior
sampler <- function(K) {
  mu    <- rnorm(K)
  sigma <- rgamma(K, 1)
  theta <- matrix(c(mu, sigma), byrow = TRUE, ncol = K)
  return(theta)
}

# Data simulator
simulator <- function(theta_set, m) {
  apply(theta_set, 2, function(theta) {
    Z <- rnorm(m, theta[1], theta[2])
    dim(Z) <- c(1, m)
    Z
  }, simplify = FALSE)
}
m <- 15


test_that("the neural estimator can be trained with fixed a training set", {
  
  theta_train <- sampler(100)
  theta_val   <- sampler(100)
  Z_train <- simulator(theta_train, m)
  Z_val   <- simulator(theta_val, m)
  
  expect_error(train(estimator, Z_train = Z_train, Z_val = Z_val, epochs = 2, verbose = F))
  
  estimator <- train(
    estimator,
    theta_train = theta_train,
    theta_val   = theta_val,
    Z_train = Z_train,
    Z_val   = Z_val,
    epochs = 2, 
    verbose = F
  )
  expect_equal(1, 1)
})

test_that("the neural estimator can be trained with simulation on-the-fly (using Julia functions)", {
  
  # Parameter sampler
  sampler <- juliaEval("
      function sampler(K)
      	mu = randn(K)
      	sigma = rand(K)
      	theta = hcat(mu, sigma)'
      	return theta
      end")
  
  # Data simulator
  simulator <- juliaEval("
      simulator(theta_matrix, m) = [theta[1] .+ theta[2] * randn(1, m) for theta in eachcol(theta_matrix)]
      ")
  
  
  estimator  <- train(estimator, sampler = sampler, simulator = simulator, m = m, epochs = 2, verbose = F)
  estimator  <- train(estimator, sampler = sampler, simulator = simulator, m = m, epochs = 2, loss = "squared-error", verbose = F)
  estimator  <- train(estimator, sampler = sampler, simulator = simulator, m = m, epochs = 2, loss = "Flux.Losses.mae", verbose = F)
  
  expect_error(train(estimator, sampler = sampler, simulator = simulator))
  expect_error(train(estimator, sampler = sampler, Z_train = Z_train, Z_val = Z_val, epochs = 2, verbose = F))
  expect_error(train(estimator, sampler = sampler, theta_train = theta_train, theta_val = theta_val, epochs = 2, verbose = F))
  expect_error(train(estimator, sampler = sampler, epochs = 2, verbose = F))
  expect_error(train(estimator, theta_train = theta_train, theta_val = theta_val, simulator = simulator, Z_train = Z_train, Z_val = Z_val, epochs = 2, verbose = F))
  
  expect_equal(1, 1)
})

test_that("the neural estimator can be assessed with assess()", {
  theta_test  <- sampler(100)
  Z_test      <- simulator(theta_test, m)
  assessment  <- assess(estimator, theta_test, Z_test)
  risk(assessment)
  bias(assessment)
  rmse(assessment)
  
  # Test that parameters can be given as a vector in single-parameter case
  estimator_one_param <- juliaEval('
    using NeuralEstimators, Flux
    p = 1    # number of parameters in the statistical model
    w = 32   # number of neurons in each layer
    psi = Chain(Dense(1, w, relu), Dense(w, w, relu), Dense(w, w, relu))
    phi = Chain(Dense(w, w, relu), Dense(w, p))
    estimator = PointEstimator(DeepSet(psi, phi))
  ')
  assessment <- assess(estimator_one_param, rnorm(100), Z_test)
  
  expect_equal(1, 1)
})

test_that("the neural estimator can be applied to real data using estimate() and bootstrap()", {
  # Generate some "observed" data
  theta    <- as.matrix(c(0, 0.5))         # true parameters
  Z        <- simulator(theta, m)          # pretend that this is observed data
  thetahat <- estimate(estimator, Z)       # point estimates
  p = 2
  expect_equal(nrow(thetahat), p)
  expect_equal(ncol(thetahat), 1)
  
  ## Non-parametric bootstrap estimates
  B  <- 400
  bs <- bootstrap(estimator, Z, B = B)  
  expect_equal(nrow(bs), p)
  expect_equal(ncol(bs), B)
})

test_that("neural ratio estimator can be constructed and used to make inference", {
  
  estimator <- juliaEval('
    using NeuralEstimators, Flux
    d = 2    # number of parameters in the statistical model
    w = 32   # number of neurons in each layer
    num_summaries = 3d
    psi = Chain(Dense(1, w, relu), Dense(w, w, relu), Dense(w, w, relu))
    phi = Chain(Dense(w, w, relu), Dense(w, num_summaries))
    summary_network = DeepSet(psi, phi)
    estimator = RatioEstimator(summary_network, d; num_summaries = num_summaries)
')
  
  theta <- as.matrix(c(0, 0.5))            # true parameters
  Z     <- simulator(theta, m)             # "observed" data
  ratio <- logratio(estimator, Z, theta)   # ratio estimate
  expect_equal(nrow(ratio), 1)
  expect_equal(ncol(ratio), 1)
  ratio <- logratio(estimator, Z, cbind(theta, theta))   # ratio estimates 
  expect_equal(nrow(ratio), 1)
  expect_equal(ncol(ratio), 2)
  
  # Grid-based methods for estimation/posterior sampling
  grid <- t(expand.grid(seq(0, 1, len = 50), seq(0, 1, len = 50)))
  sampleposterior(estimator, Z[[1]], grid = grid) 
  sampleposterior(estimator, Z, grid = grid)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missingdata.R
\name{encodedata}
\alias{encodedata}
\title{encodedata}
\usage{
encodedata(Z, c = 0)
}
\arguments{
\item{Z}{data containing \code{NA} entries}

\item{c}{fixed constant with which to replace \code{NA} entries}
}
\value{
Augmented data set (U, W). If \code{Z} is provided as a list, the return type will be a \code{JuliaProxy} object; these objects can be indexed in the usual manner using \code{[[}, or converted to an R object using \code{juliaGet()} (note however that \code{juliaGet()} can be slow for large data sets).
}
\description{
For data \code{Z} with missing (\code{NA}) entries, computes an augmented data set (U, W) where W encodes the missingness pattern as an indicator vector and U is the original data Z with missing entries replaced by a fixed constant \code{c}.
}
\examples{
\dontrun{
library("NeuralEstimators")
Z <- matrix(c(1, 2, NA, NA, 5, 6, 7, NA, 9), nrow = 3)
encodedata(Z)
encodedata(list(Z, Z))}
}
\seealso{
the Julia version of \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/miscellaneous#Utility-functions}{\code{encodedata()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{risk}
\alias{risk}
\title{computes a Monte Carlo approximation of an estimator's Bayes risk}
\usage{
risk(
  assessment,
  loss = function(x, y) abs(x - y),
  average_over_parameters = FALSE,
  average_over_sample_sizes = TRUE
)
}
\arguments{
\item{assessment}{an object returned by \code{\link[=assess]{assess()}}}

\item{loss}{a binary operator defining the loss function (default absolute-error loss)}

\item{average_over_parameters}{if \code{TRUE}, the loss is averaged over all parameters; otherwise (default), the loss is averaged over each parameter separately}

\item{average_over_sample_sizes}{if \code{TRUE} (default), the loss is averaged over all sample sizes (the column \code{m} in \code{df}); otherwise, the loss is averaged over each sample size separately}
}
\value{
a dataframe giving an estimate of the Bayes risk
}
\description{
computes a Monte Carlo approximation of an estimator's Bayes risk
}
\seealso{
\code{\link[=assess]{assess()}}, \code{\link[=bias]{bias()}}, \code{\link[=rmse]{rmse()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{savestate}
\alias{savestate}
\title{save the state of a neural estimator}
\usage{
savestate(estimator, filename)
}
\arguments{
\item{estimator}{the neural estimator that we wish to save}

\item{filename}{file in which to save the neural-network state as a \code{bson} file}
}
\value{
No return value, called for side effects
}
\description{
save the state of a neural estimator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{logratio}
\alias{logratio}
\title{logratio}
\usage{
logratio(estimator, Z, grid, batchsize = 32, ...)
}
\arguments{
\item{estimator}{a neural ratio estimator}

\item{Z}{data in a format amenable to the neural-network architecture of \code{estimator}}

\item{grid}{matrix of parameter values, where each column is a parameter configuration}

\item{batchsize}{the batch size}

\item{...}{additional keyword arguments passed to the Julia version of \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/inference#Inference-with-observed-data}{\code{logratio()}}}
}
\value{
A matrix of log ratios with one row per data set and one column per grid point
}
\description{
Apply a neural ratio estimator to data
}
\seealso{
\code{\link[=sampleposterior]{sampleposterior()}} for making posterior inferences
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missingdata.R
\name{spatialgraph}
\alias{spatialgraph}
\title{spatialgraph}
\usage{
spatialgraph(S, Z, isotropic = TRUE, stationary = TRUE, ...)
}
\arguments{
\item{S}{Spatial locations, provided as:
\itemize{
\item An \eqn{n\times 2}{n x 2} matrix when locations are fixed across replicates, where \eqn{n} is the number of spatial locations.
\item A list of \eqn{n_i \times 2}{ni x 2} matrices when locations vary across replicates.
\item A list of the above elements (i.e., a list of matrices or a list of lists of matrices) when constructing graphs from multiple data sets.
}}

\item{Z}{Spatial data, provided as:
\itemize{
\item An \eqn{n\times m}{n x m} matrix when locations are fixed, where \eqn{m} is the number of replicates.
\item A list of \eqn{n_i}{ni}-vectors when locations vary across replicates.
\item A list of the above elements (i.e., a list of matrices or a list of lists of vectors) when constructing graphs from multiple data sets.
}}

\item{isotropic}{Logical. If \code{TRUE}, edge features store the spatial distance (magnitude)
between nodes. If \code{FALSE}, the spatial displacement or spatial location is stored, depending
on the value of \code{stationary}.}

\item{stationary}{Logical. If \code{TRUE}, edge features store the spatial displacement
(vector difference) between nodes, capturing both magnitude and direction. If \code{FALSE},
edge features include the full spatial locations of both nodes.}

\item{...}{Additional keyword arguments from the Julia function \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/miscellaneous#Utility-functions}{\code{adjacencymatrix()}}
that define the neighborhood of each node, with the default being a randomly
selected set of \code{k=30} neighbors within a radius of \code{r=0.15} spatial distance units.}
}
\value{
A \code{GNNGraph} (\code{JuliaProxy} object) or, if multiple data sets are provided, a vector of \code{GNNGraph} objects which can be indexed in the usual manner using \code{[[} or converted to an R list using a combination of indexing and \code{lapply}.
}
\description{
Constructs a graph object for use in a graph neural network (GNN).
}
\examples{
\dontrun{
library("NeuralEstimators")

# Number of replicates
m <- 5

# Spatial locations fixed for all replicates
n <- 100 
S <- matrix(runif(n * 2), n, 2)
Z <- matrix(runif(n * m), n, m)
g <- spatialgraph(S, Z)

# Spatial locations varying between replicates
n <- sample(50:100, m, replace = TRUE)
S <- lapply(n, function(ni) matrix(runif(ni * 2), ni, 2))
Z <- lapply(n, function(ni) runif(ni))
g <- spatialgraph(S, Z)

# Multiple data sets: Spatial locations fixed for all replicates within a given data set
K <- 15 # number of data sets
n <- sample(50:100, K, replace = TRUE) # number of spatial locations can vary between data sets
S <- lapply(1:K, function(k) matrix(runif(n[k] * 2), n[k], 2))
Z <- lapply(1:K, function(k) matrix(runif(n[k] * m), n[k], m))
g <- spatialgraph(S, Z)

# Multiple data sets: Spatial locations varying between replicates within a given data set
S <- lapply(1:K, function(k) {
  lapply(1:m, function(i) {
  ni <- sample(50:100, 1)       # randomly generate the number of locations for each replicate
  matrix(runif(ni * 2), ni, 2)  # generate the spatial locations
  })
})
Z <- lapply(1:K, function(k) {
  lapply(1:m, function(i) {
    n <- nrow(S[[k]][[i]])
    runif(n)  
  })
})
g <- spatialgraph(S, Z)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{rmse}
\alias{rmse}
\title{computes a Monte Carlo approximation of an estimator's root-mean-square error (RMSE)}
\usage{
rmse(assessment, ...)
}
\arguments{
\item{assessment}{an object returned by \code{\link[=assess]{assess()}}}

\item{...}{optional arguments inherited from \code{risk()} (excluding the argument \code{loss})}
}
\value{
a dataframe giving the estimated RMSE
}
\description{
computes a Monte Carlo approximation of an estimator's root-mean-square error (RMSE)
}
\seealso{
\code{\link[=assess]{assess()}}, \code{\link[=bias]{bias()}}, \code{\link[=risk]{risk()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NeuralEstimators.R
\docType{package}
\name{NeuralEstimators-package}
\alias{NeuralEstimators}
\alias{NeuralEstimators-package}
\title{NeuralEstimators: Likelihood-Free Parameter Estimation using Neural Networks}
\description{
An 'R' interface to the 'Julia' package 'NeuralEstimators.jl'. The package facilitates the user-friendly development of neural Bayes estimators, which are neural networks that map data to a point summary of the posterior distribution (Sainsbury-Dale et al., 2024, \doi{10.1080/00031305.2023.2249522}). These estimators are likelihood-free and amortised, in the sense that, once the neural networks are trained on simulated data, inference from observed data can be made in a fraction of the time required by conventional approaches. The package also supports amortised Bayesian or frequentist inference using neural networks that approximate the posterior or likelihood-to-evidence ratio (Zammit-Mangion et al., 2025, Sec. 3.2, 5.2, \doi{10.48550/arXiv.2404.12484}). The package accommodates any model for which simulation is feasible by allowing users to define models implicitly through simulated data.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/msainsburydale/NeuralEstimators}
  \item \url{https://msainsburydale.github.io/NeuralEstimators.jl/dev/}
}

}
\author{
\strong{Maintainer}: Matthew Sainsbury-Dale \email{msainsburydale@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{bootstrap}
\alias{bootstrap}
\title{bootstrap}
\usage{
bootstrap(estimator, Z, B = 400, blocks = NULL, use_gpu = TRUE)
}
\arguments{
\item{estimator}{a neural point estimator}

\item{Z}{either a list of data sets simulated conditionally on the fitted parameters (parametric bootstrap); or a single observed data set containing independent replicates, which will be sampled with replacement \code{B} times (non-parametric bootstrap)}

\item{B}{number of non-parametric bootstrap samples}

\item{blocks}{integer vector specifying the blocks in non-parameteric bootstrap. For example, with 5 replicates, the first two corresponding to block 1 and the remaining three corresponding to block 2, \code{blocks} should be \code{c(1,1,2,2,2)}}

\item{use_gpu}{boolean indicating whether to use the GPU if it is available}
}
\value{
d × \code{B} matrix, where d is the dimension of the parameter vector
}
\description{
Compute bootstrap estimates from a neural point estimator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{bias}
\alias{bias}
\title{computes a Monte Carlo approximation of an estimator's bias}
\usage{
bias(assessment, ...)
}
\arguments{
\item{assessment}{an object returned by \code{\link[=assess]{assess()}}}

\item{...}{optional arguments inherited from \code{risk()} (excluding the argument \code{loss})}
}
\value{
a dataframe giving the estimated bias
}
\description{
computes a Monte Carlo approximation of an estimator's bias
}
\seealso{
\code{\link[=assess]{assess()}}, \code{\link[=risk]{risk()}}, \code{\link[=rmse]{rmse()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{loadstate}
\alias{loadstate}
\title{Load a saved state of a neural estimator}
\usage{
loadstate(estimator, filename)
}
\arguments{
\item{estimator}{the neural estimator that we wish to load the state into}

\item{filename}{file name (including path) of the neural-network state stored in a \code{bson} file}
}
\value{
\code{estimator} updated with the saved state
}
\description{
Load a saved state of a neural estimator (e.g., optimised neural-network parameters). Useful for amortised inference, whereby a neural network is trained once and then used repeatedly to make inference with new data sets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{assess}
\alias{assess}
\title{assess a neural estimator}
\usage{
assess(estimator, parameters, Z, ...)
}
\arguments{
\item{estimator}{a neural estimator (or a list of neural estimators)}

\item{parameters}{true parameters, stored as a \eqn{d\times K}{dxK} matrix, where \eqn{d} is the dimension of the parameter vector and \eqn{K} is the number of sampled parameter vectors}

\item{Z}{data simulated conditionally on the \code{parameters}. If \code{length(Z)} > K, the parameter matrix will be recycled by horizontal concatenation as \code{parameters = parameters[, rep(1:K, J)]}, where \code{J = length(Z) / K}}

\item{...}{additional keyword arguments passed to the Julia version of \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/assessment}{\code{assess()}}}
}
\value{
a list of two data frames: \code{runtimes} contains the
total time taken for each estimator, while \code{df} is a long-form
data frame with columns:
\itemize{
\item{"parameter"; the name of the parameter}
\item{"truth"; the true value of the parameter}
\item{"estimate"; the estimated value of the parameter}
\item{"k"; the index of the parameter vector in the test set}
\item{"j"; the index of the data set}
}
}
\description{
assess a neural estimator
}
\seealso{
\code{\link[=risk]{risk()}}, \code{\link[=rmse]{rmse()}}, \code{\link[=bias]{bias()}}, \code{\link[=plotestimates]{plotestimates()}}, and \code{\link[=plotdistribution]{plotdistribution()}} for computing various empirical diagnostics and visualisations from an object returned by \code{assess()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{sampleposterior}
\alias{sampleposterior}
\title{sampleposterior}
\usage{
sampleposterior(estimator, Z, N = 1000, ...)
}
\arguments{
\item{estimator}{a neural posterior or likelihood-to-evidence-ratio estimator}

\item{Z}{data in a format amenable to the neural-network architecture of \code{estimator}}

\item{N}{number of approximate posterior samples to draw}

\item{...}{additional keyword arguments passed to the Julia version of \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/inference#Inference-with-observed-data}{\code{sampleposterior()}}}
}
\value{
d × \code{N} matrix of posterior samples, where d is the dimension of the parameter vector. If \code{Z} contains multiple independent data sets, a list of matrices will be returned
}
\description{
Samples from the approximate posterior distribution given data \code{Z}.
}
\seealso{
\code{\link[=estimate]{estimate()}} for making inference with neural Bayes estimators
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{train}
\alias{train}
\title{Train a neural estimator}
\usage{
train(
  estimator,
  sampler = NULL,
  simulator = NULL,
  theta_train = NULL,
  theta_val = NULL,
  Z_train = NULL,
  Z_val = NULL,
  K = 10000,
  m = NULL,
  sampler_args = NULL,
  sampler_kwargs = NULL,
  simulator_args = NULL,
  simulator_kwargs = NULL,
  loss = "absolute-error",
  learning_rate = 5e-04,
  epochs = 100,
  batchsize = 32,
  savepath = NULL,
  stopping_epochs = 5,
  epochs_per_Z_refresh = 1,
  epochs_per_theta_refresh = 1,
  simulate_just_in_time = FALSE,
  use_gpu = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{estimator}{a neural estimator}

\item{sampler}{a function that takes an integer \code{K}, samples \code{K} parameter vectors from the prior, and returns them as a px\code{K} matrix}

\item{simulator}{a function that takes a px\code{K} matrix of parameters and an integer \code{m}, and returns \code{K} simulated data sets each containing \code{m} independent replicates}

\item{theta_train}{a set of parameters used for updating the estimator using stochastic gradient descent}

\item{theta_val}{a set of parameters used for monitoring the performance of the estimator during training}

\item{Z_train}{a simulated data set used for updating the estimator using stochastic gradient descent}

\item{Z_val}{a simulated data set used for monitoring the performance of the estimator during training}

\item{K}{the number of parameter vectors sampled in the training set at each epoch; the size of the validation set is set to \code{K}/5.}

\item{m}{deprecated; use \code{simulator_args}}

\item{sampler_args}{a list of positional arguments passed to the parameter sampler; if provided, \code{sampler} is called as \code{sampler(K, sampler_args...)}.}

\item{sampler_kwargs}{a list of positional arguments passed to the parameter sampler; if provided, \code{sampler} is called as \code{sampler(K; sampler_kwargs...)}.}

\item{simulator_args}{a list of positional arguments passed to the data simulator; if provided, \code{simulator} is called as \code{simulator(theta, simulator_args...)}.}

\item{simulator_kwargs}{a list of positional arguments passed to the data simulator; if provided, \code{simulator} is called as \code{simulator(theta; simulator_kwargs...)}.}

\item{loss}{the loss function: a string ('absolute-error' for mean-absolute-error loss or 'squared-error' for mean-squared-error loss), or a string of Julia code defining the loss function. For some classes of estimators (e.g., \code{PosteriorEstimator}, \code{QuantileEstimator}, \code{RatioEstimator}), the loss function does not need to be specified.}

\item{learning_rate}{the initial learning rate for the optimiser ADAM (default 5e-4)}

\item{epochs}{the number of epochs to train the neural network. An epoch is one complete pass through the entire training data set when doing stochastic gradient descent.}

\item{batchsize}{the batchsize to use when performing stochastic gradient descent, that is, the number of training samples processed between each update of the neural-network parameters.}

\item{savepath}{path to save the trained estimator and other information; if null (default), nothing is saved. Otherwise, the neural-network parameters (i.e., the weights and biases) will be saved during training as \code{bson} files; the risk function evaluated over the training and validation sets will also be saved, in the first and second columns of \code{loss_per_epoch.csv}, respectively; the best parameters (as measured by validation risk) will be saved as \code{best_network.bson}.}

\item{stopping_epochs}{cease training if the risk doesn't improve in this number of epochs (default 5).}

\item{epochs_per_Z_refresh}{integer indicating how often to refresh the training data}

\item{epochs_per_theta_refresh}{integer indicating how often to refresh the training parameters; must be a multiple of \code{epochs_per_Z_refresh}}

\item{simulate_just_in_time}{flag indicating whether we should simulate "just-in-time", in the sense that only a \code{batchsize} number of parameter vectors and corresponding data are in memory at a given time}

\item{use_gpu}{a boolean indicating whether to use the GPU if one is available}

\item{verbose}{a boolean indicating whether information, including empirical risk values and timings, should be printed to the console during training.}
}
\value{
a trained neural estimator or, if \code{m} is a vector, a list of trained neural estimators
}
\description{
The function caters for different variants of "on-the-fly" simulation.
Specifically, a \code{sampler} can be provided to continuously sample new
parameter vectors from the prior, and a \code{simulator} can be provided to
continuously simulate new data conditional on the parameters. If provided
with specific sets of parameters (\code{theta_train} and \code{theta_val})
and/or data (\code{Z_train} and \code{Z_val}), they will be held fixed during
training.

Note that using \code{R} functions to perform "on-the-fly" simulation requires the user to have installed the Julia package \code{RCall}.
}
\examples{
\dontrun{
# Construct a neural Bayes estimator for replicated univariate Gaussian 
# data with unknown mean and standard deviation. 

# Load R and Julia packages
library("NeuralEstimators")
library("JuliaConnectoR")
juliaEval("using NeuralEstimators, Flux")

# Define the neural-network architecture
estimator <- juliaEval('
 n = 1    # dimension of each replicate
 d = 2    # number of parameters in the model
 w = 32   # width of each hidden layer
 psi = Chain(Dense(n, w, relu), Dense(w, w, relu))
 phi = Chain(Dense(w, w, relu), Dense(w, d))
 deepset = DeepSet(psi, phi)
 estimator = PointEstimator(deepset)
')

# Sampler from the prior
sampler <- function(K) {
  mu    <- rnorm(K)      # Gaussian prior for the mean
  sigma <- rgamma(K, 1)  # Gamma prior for the standard deviation
  theta <- matrix(c(mu, sigma), byrow = TRUE, ncol = K)
  return(theta)
}

# Data simulator
simulator <- function(theta_set, m) {
  apply(theta_set, 2, function(theta) {
    t(rnorm(m, theta[1], theta[2]))
  }, simplify = FALSE)
}

# Train using fixed parameter and data sets 
theta_train <- sampler(10000)
theta_val   <- sampler(2000)
m <- 30 # number of iid replicates
Z_train <- simulator(theta_train, m)
Z_val   <- simulator(theta_val, m)
estimator <- train(estimator, 
                   theta_train = theta_train, 
                   theta_val = theta_val, 
                   Z_train = Z_train, 
                   Z_val = Z_val)
                   
##### Simulation on-the-fly using R functions ####

juliaEval("using RCall") # requires the Julia package RCall
estimator <- train(estimator, sampler = sampler, simulator = simulator, m = m)

##### Simulation on-the-fly using Julia functions ####

# Defining the sampler and simulator in Julia can improve computational 
# efficiency by avoiding the overhead of communicating between R and Julia. 

juliaEval("using Distributions")

# Parameter sampler
sampler <- juliaEval("
      function sampler(K)
      	mu = rand(Normal(0, 1), K)
      	sigma = rand(Gamma(1), K)
      	theta = hcat(mu, sigma)'
      	return theta
      end")

# Data simulator
simulator <- juliaEval("
      function simulator(theta_matrix, m)
      	Z = [rand(Normal(theta[1], theta[2]), 1, m) for theta in eachcol(theta_matrix)]
      	return Z
      end")

# Train
estimator <- train(estimator, sampler = sampler, simulator = simulator, m = m)}
}
\seealso{
the Julia version of \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/training}{\code{train()}}, \code{\link[=assess]{assess()}} for assessing an estimator post training, and \code{\link[=estimate]{estimate()}}/\code{\link[=sampleposterior]{sampleposterior()}} for making inference with observed data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{estimate}
\alias{estimate}
\title{estimate}
\usage{
estimate(estimator, Z, X = NULL, batchsize = 32, ...)
}
\arguments{
\item{estimator}{a neural estimator that can be applied to data in a call of the form \code{estimator(Z)}}

\item{Z}{data in a format amenable to the neural-network architecture of \code{estimator}}

\item{X}{additional inputs to the neural network; if provided, the call will be of the form \verb{estimator((Z, X))}}

\item{batchsize}{the batch size for applying \code{estimator} to \code{Z}}

\item{...}{additional keyword arguments passed to the Julia version of \href{https://msainsburydale.github.io/NeuralEstimators.jl/dev/API/inference#Inference-with-observed-data}{\code{estimate()}}}
}
\value{
a matrix of outputs resulting from applying \code{estimator} to \code{Z} (and possibly \code{X})
}
\description{
Apply a neural Bayes estimator to data
}
\seealso{
\code{\link[=sampleposterior]{sampleposterior()}} for making inference with neural posterior or likelihood-to-evidence-ratio estimators
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plotestimates}
\alias{plotestimates}
\title{Plot estimates vs. true values.}
\usage{
plotestimates(
  df,
  estimator_labels = ggplot2::waiver(),
  parameter_labels = NULL
)
}
\arguments{
\item{df}{a long form data frame containing fields \code{estimator}, \code{parameter}, \code{estimate}, and \code{truth}.}

\item{estimator_labels}{a named vector containing estimator labels.}

\item{parameter_labels}{a named vector containing parameter labels.}
}
\value{
a \code{'ggplot'} of the estimates for each parameter against the true value.
}
\description{
Plot estimates vs. true values.
}
\examples{
\dontrun{
K <- 50
df <- data.frame(
  estimator = c("Estimator 1", "Estimator 2"), 
  parameter = rep(c("mu", "sigma"), each = K),
  truth = 1:(2*K), 
  estimate = 1:(2*K) + rnorm(4*K)
)
estimator_labels <- c("Estimator 1" = expression(hat(theta)[1]("·")),
                      "Estimator 2" = expression(hat(theta)[2]("·")))
parameter_labels <- c("mu" = expression(mu), "sigma" = expression(sigma))

plotestimates(df,  parameter_labels = parameter_labels, estimator_labels)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missingdata.R
\name{tanhloss}
\alias{tanhloss}
\title{tanhloss}
\usage{
tanhloss(k)
}
\arguments{
\item{k}{Positive numeric value that controls the smoothness of the approximation.}
}
\value{
String defining the tanh loss function in Julia code.
}
\description{
For \code{k} > 0, defines Julia code that defines the loss function,
\deqn{L(\hat{\theta}, \theta) = \tanh\left(\frac{|\hat{\theta} - \theta|}{k}\right),}
which approximates the 0-1 loss as \code{k} tends to zero.

The resulting string is intended to be used in the function \code{\link{train}}, but can also be converted to a callable function using \code{juliaEval}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plotdistribution}
\alias{plotdistribution}
\title{Plot the empirical sampling distribution of an estimator.}
\usage{
plotdistribution(
  df,
  type = c("box", "density", "scatter"),
  parameter_labels = NULL,
  estimator_labels = ggplot2::waiver(),
  truth_colour = "black",
  truth_size = 8,
  truth_line_size = NULL,
  pairs = FALSE,
  upper_triangle_plots = NULL,
  legend = TRUE,
  return_list = FALSE,
  flip = FALSE
)
}
\arguments{
\item{df}{a long form data frame containing fields \code{estimator}, \code{parameter}, \code{estimate}, \code{truth}, and a column (e.g., \code{replicate}) to uniquely identify each observation.}

\item{type}{string indicating whether to plot kernel density estimates for each individual parameter (\code{type = "density"}) or scatter plots for all parameter pairs (\code{type = "scatter"}).}

\item{parameter_labels}{a named vector containing parameter labels.}

\item{estimator_labels}{a named vector containing estimator labels.}

\item{truth_colour}{the colour used to denote the true parameter value.}

\item{truth_size}{the size of the point used to denote the true parameter value (applicable only for \code{type = "scatter"}).}

\item{truth_line_size}{the size of the cross-hairs used to denote the true parameter value. If \code{NULL} (default), the cross-hairs are not plotted. (applicable only for \code{type = "scatter"}).}

\item{pairs}{logical; should we combine the scatter plots into a single pairs plot (applicable only for \code{type = "scatter"})?}

\item{upper_triangle_plots}{an optional list of plots to include in the uppertriangle of the pairs plot.}

\item{legend}{Flag; should we include the legend (only applies when constructing a pairs plot)}

\item{return_list}{Flag; should the parameters be split into a list?}

\item{flip}{Flag; should the boxplots be "flipped" using \code{coord_flip()} (default \code{FALSE})?}
}
\value{
a list of \code{'ggplot'} objects or, if \code{pairs = TRUE}, a single \code{'ggplot'}.
}
\description{
Plot the empirical sampling distribution of an estimator.
}
\examples{
\dontrun{
# In the following, we have two estimators and, for each parameter, 50 estimates
# from each estimator.

estimators <- c("Estimator 1", "Estimator 2")
estimator_labels <- c("Estimator 1" = expression(hat(theta)[1]("·")),
                      "Estimator 2" = expression(hat(theta)[2]("·")))

# Single parameter:
df <- data.frame(
  estimator = estimators, truth = 0, parameter = "mu",
  estimate  = rnorm(2*50),
  replicate = rep(1:50, each = 2)
)
parameter_labels <- c("mu" = expression(mu))
plotdistribution(df)
plotdistribution(df, type = "density")
plotdistribution(df, parameter_labels = parameter_labels, estimator_labels = estimator_labels)

# Two parameters:
df <- rbind(df, data.frame(
  estimator = estimators, truth = 1, parameter = "sigma",
  estimate  = rgamma(2*50, shape = 1, rate = 1),
  replicate = rep(1:50, each = 2)
))
parameter_labels <- c(parameter_labels, "sigma" = expression(sigma))
plotdistribution(df, parameter_labels = parameter_labels)
plotdistribution(df, parameter_labels = parameter_labels, type = "density")
plotdistribution(df, parameter_labels = parameter_labels, type = "scatter")

# Three parameters:
df <- rbind(df, data.frame(
  estimator = estimators, truth = 0.25, parameter = "alpha",
  estimate  = 0.5 * runif(2*50),
  replicate = rep(1:50, each = 2)
))
parameter_labels <- c(parameter_labels, "alpha" = expression(alpha))
plotdistribution(df, parameter_labels = parameter_labels)
plotdistribution(df, parameter_labels = parameter_labels, type = "density")
plotdistribution(df, parameter_labels = parameter_labels, type = "scatter")
plotdistribution(df, parameter_labels = parameter_labels, type = "scatter", pairs = TRUE)

# Pairs plot with user-specified plots in the upper triangle:
upper_triangle_plots <- lapply(1:3, function(i) {
  x = rnorm(10)
  y = rnorm(10)
  shape = sample(c("Class 1", "Class 2"), 10, replace = TRUE)
  ggplot() +
    geom_point(aes(x = x, y = y, shape = shape)) + 
    labs(shape = "") +
    theme_bw()
})
plotdistribution(
    df, 
    parameter_labels = parameter_labels, estimator_labels = estimator_labels,
    type = "scatter", pairs = TRUE, upper_triangle_plots = upper_triangle_plots
    )}
}
